from __future__ import annotations

from typing import Literal

from playwright.sync_api import Locator, Page
from playwright.sync_api import expect as playwright_expect

from .._types import PatternOrStr, Timeout
from ._base import UiBase


class Chat(UiBase):
    """Controller for :func:`shiny.ui.chat`."""

    loc: Locator
    """
    Playwright `Locator` for the chat.
    """
    loc_messages: Locator
    """
    Playwright `Locator` for the chat messages.
    """
    loc_latest_message: Locator
    """
    Playwright `Locator` for the last message in the chat.
    """
    loc_input_container: Locator
    """
    Playwright `Locator` for the chat input container.
    """
    loc_input: Locator
    """
    Playwright `Locator` for the chat's <textarea> input.
    """
    loc_input_button: Locator
    """
    Playwright `Locator` for the chat's <button> input.
    """

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a new instance of the `Chat` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the chat.
        """
        super().__init__(
            page,
            id=id,
            loc=f"#{id}",
        )
        self.loc_messages = self.loc.locator("> shiny-chat-messages")
        self.loc_latest_message = self.loc_messages.locator("> :last-child")
        self.loc_input_container = self.loc.locator("> shiny-chat-input")
        self.loc_input = self.loc_input_container.locator("textarea")
        self.loc_input_button = self.loc_input_container.locator("button")

    def expect_latest_message(
        self,
        value: PatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the last message in the chat.

        Parameters
        ----------
        value
            The expected last message.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        # playwright_expect(self.loc_latest_message).to_have_text(value, timeout=timeout)
        playwright_expect(self.loc_latest_message).to_have_text(
            value, use_inner_text=True, timeout=timeout
        )

    def expect_messages(
        self,
        value: PatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the chat messages.

        Parameters
        ----------
        value
            The expected messages.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        playwright_expect(self.loc_messages).to_have_text(
            value, use_inner_text=True, timeout=timeout
        )

    def set_user_input(
        self,
        value: str,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Sets the user message in the chat.

        Parameters
        ----------
        value
            The message to send.
        timeout
            The maximum time to wait for the chat input to be visible and interactable. Defaults to `None`.
        """
        self.loc_input.type(value, timeout=timeout)

    def send_user_input(
        self, *, method: Literal["enter", "click"] = "enter", timeout: Timeout = None
    ) -> None:
        """
        Sends the user message in the chat.

        Parameters
        ----------
        method
            The method to send the user message. Defaults to `"enter"`.
        timeout
            The maximum time to wait for the chat input to be visible and interactable. Defaults to `None`.
        """
        if method == "enter":
            self.loc_input.press("Enter", timeout=timeout)
        else:
            self.loc_input_button.click(timeout=timeout)

    def expect_user_input(
        self, value: PatternOrStr, *, timeout: Timeout = None
    ) -> None:
        """
        Expects the user message in the chat.

        Parameters
        ----------
        value
            The expected user message.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        playwright_expect(self.loc_input).to_have_value(value, timeout=timeout)



import re

from conftest import create_doc_example_core_fixture
from playwright.sync_api import Locator, Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc

app = create_doc_example_core_fixture("input_text_area")

default_txt = "Data summary\nwith\nmultiple\nlines"


def test_input_text_area_kitchen(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    caption = controller.InputTextArea(page, "caption_regular")
    caption.expect.to_have_value(default_txt)
    expect(caption.loc).to_have_value(default_txt)

    expect(caption.loc_label).to_have_text("Caption:")

    # Better approach
    expect(caption.loc_label).to_have_text("Caption:")
    expect(caption.loc).to_have_value(default_txt)
    expect(caption.loc).not_to_have_attribute("width", re.compile(r".*"))
    expect(caption.loc).not_to_have_attribute("placeholder", re.compile(r".*"))
    expect(caption.loc).not_to_have_attribute("autocomplete", re.compile(r".*"))
    expect(caption.loc).not_to_have_attribute("spellcheck", re.compile(r".*"))

    # Best approach
    caption.expect_label("Caption:")
    caption.expect_value("Data summary\nwith\nmultiple\nlines")
    caption.expect_width(None)
    caption.expect_placeholder(None)
    caption.expect_autocomplete(None)
    caption.expect_spellcheck(None)
    caption.expect_autoresize(False)

    controller.InputTextArea(page, "caption_autoresize").expect_autoresize(True)


def test_input_text_typical(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    caption = controller.InputTextArea(page, "caption_regular")
    caption.expect.to_have_value(default_txt)
    caption.set("test value")
    caption.expect.not_to_have_value(default_txt)
    caption.expect.to_have_value("test value")


def test_input_text_app(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    caption = controller.InputTextArea(page, "caption_regular")
    caption.expect.to_have_value(default_txt)

    value = controller.OutputTextVerbatim(page, "value_regular")
    value.expect_value(default_txt)

    caption.set("test value")
    caption.expect.to_have_value("test value")
    value.expect_value("test value")


resize_number = 6


def get_box_height(locator: Locator) -> float:
    bounding_box = locator.bounding_box()
    if bounding_box is not None:
        return bounding_box["height"]
    else:
        return 0


def test_autoresize(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    input_area = controller.InputTextArea(page, "caption_autoresize")
    output_txt_verbatim = controller.OutputTextVerbatim(page, "value_autoresize")
    input_area.set("test value")
    # use bounding box approach since height is dynamic
    initial_height = get_box_height(input_area.loc)
    output_txt_verbatim.expect_value("test value")
    for _ in range(resize_number):
        input_area.loc.press("Enter")
    input_area.loc.type("end value")
    return_txt = "\n" * resize_number
    output_txt_verbatim.expect_value(f"test value{return_txt}end value")
    assert get_box_height(input_area.loc) > initial_height



# ------------------------------------------------------------------------------------
# A basic Shiny Chat powered by Anthropic's Claude model with Bedrock.
# To run it, you'll need an AWS Bedrock configuration.
# To get started, follow the instructions at https://aws.amazon.com/bedrock/claude/
# as well as https://github.com/anthropics/anthropic-sdk-python#aws-bedrock
# ------------------------------------------------------------------------------------
from anthropic import AnthropicBedrock
from app_utils import load_dotenv

from shiny.express import ui

# Either explicitly set the AWS environment variables before launching the app, or set
# them in a file named `.env`. The `python-dotenv` package will load `.env` as
# environment variables which can be read by `os.getenv()`.
load_dotenv()
llm = AnthropicBedrock(
    # aws_secret_key=os.getenv("AWS_SECRET_KEY"),
    # aws_access_key=os.getenv("AWS_ACCESS_KEY"),
    # aws_region=os.getenv("AWS_REGION"),
    # aws_account_id=os.getenv("AWS_ACCOUNT_ID"),
)

# Set some Shiny page options
ui.page_opts(
    title="Hello Anthropic Claude Chat",
    fillable=True,
    fillable_mobile=True,
)

# Create and display empty chat
chat = ui.Chat(id="chat")
chat.ui()


# Define a callback to run when the user submits a message
@chat.on_user_submit
async def _():
    # Get messages currently in the chat
    messages = chat.messages(format="anthropic")
    # Create a response message stream
    response = llm.messages.create(
        model="anthropic.claude-3-5-sonnet-20241022-v2:0",
        messages=messages,
        stream=True,
        max_tokens=1000,
    )
    # Append the response stream into the chat
    await chat.append_message_stream(response)



import re

from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_table_data_support(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    controller.OutputTable(page, "nw_table").expect_nrow(2)
    controller.OutputCode(page, "nw_df_type").expect_value(re.compile("narwhals"))

    controller.OutputTable(page, "md_table").expect_nrow(2)
    controller.OutputCode(page, "md_df_type").expect_value(re.compile("modin"))



"""Generates synthetic data"""

import numpy as np


def cpu_count(logical: bool = True):
    return 8 if logical else 4


last_sample = np.random.uniform(0, 100, size=cpu_count(True))


def cpu_percent(percpu: bool = False):
    global last_sample
    delta = np.random.normal(scale=10, size=len(last_sample))
    last_sample = (last_sample + delta).clip(0, 100)
    if percpu:
        return last_sample.tolist()
    else:
        return last_sample.mean()



import re

from playwright.sync_api import Page
from utils.deploy_utils import skip_on_python_version

from shiny.playwright import controller
from shiny.run import ShinyAppProc


@skip_on_python_version(
    "3.9",
    reason="Astropy (and numpy) hav difficulty loading on python 3.9. posit-dev/py-shiny#1678",
)
def test_data_frame_pandas_compatible(
    page: Page,
    local_app: ShinyAppProc,
) -> None:
    page.goto(local_app.url)

    controller.OutputCode(page, "code_original").expect_value(re.compile("astropy"))
    controller.OutputCode(page, "code_astropy").expect_value(re.compile("pandas"))
    controller.OutputCode(page, "code_data").expect_value(re.compile("pandas"))



from __future__ import annotations

from playwright.sync_api import Page

from ._base import InputActionBase, WidthLocM


# TODO: Use mixin for dowloadlink and download button
class DownloadLink(InputActionBase):
    """
    Controller for :func:`shiny.ui.download_link`.
    """

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a new instance of the `DownloadLink` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the download link.
        """
        super().__init__(
            page,
            id=id,
            loc=f"#{id}.shiny-download-link:not(.btn)",
        )


class DownloadButton(
    WidthLocM,
    InputActionBase,
):
    """
    Controller for :func:`shiny.ui.download_button`
    """

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a new instance of the `DownloadButton` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the download button.
        """
        super().__init__(
            page,
            id=id,
            loc=f"#{id}.btn.shiny-download-link",
        )



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Dict,
    Literal,
    Optional,
    Protocol,
    Tuple,
    Union,
    cast,
    runtime_checkable,
)

import narwhals.stable.v1 as nw
from htmltools import TagNode
from narwhals.stable.v1.dtypes import DType as DType
from narwhals.stable.v1.typing import DataFrameT as DataFrameT
from narwhals.stable.v1.typing import IntoDataFrame as IntoDataFrame
from narwhals.stable.v1.typing import IntoDataFrameT as IntoDataFrameT
from narwhals.stable.v1.typing import IntoExpr as IntoExpr

from ..._typing_extensions import Annotated, NotRequired, Required, TypedDict
from ...types import Jsonifiable, JsonifiableDict, ListOrTuple

if TYPE_CHECKING:
    import pandas as pd

    from ...session._utils import RenderedDeps

__all__ = (
    "IntoExpr",
    "DataFrame",
    "DataFrameT",
    "DType",
    "IntoDataFrame",
    "IntoDataFrameT",
    "PandasCompatible",
    "CellHtml",
    "ColumnSort",
    "ColumnFilterStr",
    "ColumnFilterNumber",
    "ColumnFilter",
    "DataViewInfo",
    "FrameRenderPatchInfo",
    "FrameRenderSelectionModes",
    "FrameRender",
    "frame_render_to_jsonifiable",
    "FrameJsonOptions",
    "FrameJson",
    "RowsList",
    "ColsList",
    "FrameDtypeSubset",
    "FrameDtypeCategories",
    "FrameDtype",
    "StyleInfoBody",
    "StyleInfo",
    "BrowserStyleInfoBody",
    "BrowserStyleInfo",
    "CellValue",
    "CellPatch",
    "CellPatchProcessed",
)


# ---------------------------------------------------------------------

DataFrame = nw.DataFrame
Series = nw.Series

# ---------------------------------------------------------------------


@runtime_checkable
class PandasCompatible(Protocol):
    # Signature doesn't matter, runtime_checkable won't look at it anyway
    def to_pandas(self) -> pd.DataFrame: ...


class CellHtml(TypedDict):
    isShinyHtml: bool
    obj: RenderedDeps


# ---------------------------------------------------------------------


class ColumnSort(TypedDict):
    col: int
    desc: bool


class ColumnFilterStr(TypedDict):
    col: int
    value: str


class ColumnFilterNumber(TypedDict):
    col: int
    value: (
        tuple[int | float, int | float]
        | tuple[int | float, None]
        | tuple[None, int | float]
        | Annotated[list[int | float | None], 2]
    )


ColumnFilter = Union[ColumnFilterStr, ColumnFilterNumber]


class DataViewInfo(TypedDict):
    sort: tuple[ColumnSort, ...]
    filter: tuple[ColumnFilter, ...]

    rows: tuple[int, ...]  # sorted and filtered row number
    selected_rows: tuple[int, ...]  # selected and sorted and filtered row number
    # selected_columns: tuple[int, ...]  # selected and sorted and filtered row number


# ---------------------------------------------------------------------


class FrameRenderPatchInfo(TypedDict):
    key: str


class FrameRenderSelectionModes(TypedDict):
    row: Literal["single", "multiple", "none"]
    col: Literal["single", "multiple", "none"]
    rect: Literal["cell", "region", "none"]


class FrameRender(TypedDict):
    payload: FrameJson
    patchInfo: FrameRenderPatchInfo
    selectionModes: FrameRenderSelectionModes


def frame_render_to_jsonifiable(frame_render: FrameRender) -> JsonifiableDict:
    return cast(JsonifiableDict, dict(frame_render))


# ---------------------------------------------------------------------


class FrameJsonOptions(TypedDict):
    width: NotRequired[str | float | None]
    height: NotRequired[str | float | None]
    summary: NotRequired[bool | str]
    filters: NotRequired[bool]
    editable: NotRequired[bool]
    style: NotRequired[str]
    fill: NotRequired[bool]
    styles: NotRequired[list[BrowserStyleInfo]]


class FrameJson(TypedDict):
    columns: Required[list[str]]  # column names
    # index: Required[list[Any]]  # pandas index values
    data: Required[list[list[Jsonifiable]]]  # each entry is a row of len(columns)
    typeHints: Required[
        list[FrameDtype]
    ]  # each entry is a hint for the type of the column
    options: NotRequired[FrameJsonOptions]
    htmlDeps: NotRequired[list[JsonifiableDict]]


RowsList = Optional[ListOrTuple[int]]
ColsList = Optional[ListOrTuple[Union[str, int]]]


# ---------------------------------------------------------------------


class FrameDtypeSubset(TypedDict):
    type: Literal[
        "string",
        "numeric",
        "boolean",
        "date",
        "datetime",
        "duration",
        "object",
        "unknown",
        "html",
    ]


class FrameDtypeCategories(TypedDict):
    type: Literal["categorical"]
    categories: list[str]


FrameDtype = Union[
    FrameDtypeSubset,
    FrameDtypeCategories,
]


# ---------------------------------------------------------------------

# great_tables's StyleInfo
# @dataclass(frozen=True)
# class StyleInfo:
#     locname: str
#     locnum: int
#     grpname: str | None = None
#     colname: str | None = None
#     rownum: int | None = None
#     colnum: int | None = None
#     styles: list[CellStyle] = field(default_factory=list)


# https://typing.readthedocs.io/en/latest/spec/typeddict.html#alternative-syntax
# Use alternative syntax for TypedDict to avoid key error with `class`:
StyleInfoBody = TypedDict(
    "StyleInfoBody",
    {
        "location": NotRequired[Literal["body"]],
        "rows": NotRequired[Union[int, ListOrTuple[int], ListOrTuple[bool], None]],
        "cols": NotRequired[
            Union[str, int, ListOrTuple[str], ListOrTuple[int], ListOrTuple[bool], None]
        ],
        "style": NotRequired[Union[Dict[str, Jsonifiable], None]],
        "class": NotRequired[Union[str, None]],
    },
)
StyleInfo = StyleInfoBody


BrowserStyleInfoBody = TypedDict(
    "BrowserStyleInfoBody",
    {
        "location": Required[Literal["body"]],
        "rows": Required[Union[Tuple[int, ...], None]],
        "cols": Required[Union[Tuple[int, ...], None]],
        "style": Required[Union[Dict[str, Jsonifiable], None]],
        "class": Required[Union[str, None]],
    },
)
BrowserStyleInfo = BrowserStyleInfoBody


# Cell patches ----------------------------------------------------------

# CellValue = str | TagList | Tag | HTML
CellValue = TagNode


class CellPatch(TypedDict):
    row_index: int
    column_index: int
    value: CellValue


class CellPatchProcessed(TypedDict):
    row_index: int
    column_index: int
    value: str | CellHtml
    # prev_value: CellValue


def cell_patch_processed_to_jsonifiable(
    cell_patch_processed: CellPatchProcessed,
) -> JsonifiableDict:
    return cast(JsonifiableDict, dict(cell_patch_processed))



from __future__ import annotations

import warnings
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Literal, Optional

from htmltools import (
    HTML,
    Tag,
    TagAttrs,
    TagAttrValue,
    TagChild,
    TagList,
    css,
    div,
    tags,
)

from .._docstring import add_example, no_example
from .._namespaces import ResolvedId, resolve_id_or_none
from .._typing_extensions import TypedDict
from .._utils import private_random_id
from ..session import require_active_session
from ..types import MISSING, MISSING_TYPE
from ._card import CardItem
from ._html_deps_shinyverse import components_dependencies
from ._tag import consolidate_attrs, trinary
from ._utils import css_no_sub
from .css import CssUnit, as_css_padding, as_css_unit
from .fill import as_fill_item, as_fillable_container

if TYPE_CHECKING:
    from ..session import Session

__all__ = (
    "Sidebar",
    "SidebarOpen",
    "sidebar",
    "layout_sidebar",
    "update_sidebar",
)

SidebarOpenValue = Literal["open", "closed", "always"]
"""
A possible value for the `open` parameter in :func:`~shiny.ui.sidebar`:

* `"open"`: the sidebar starts open
* `"closed"`: the sidebar starts closed
* `"always"`: the sidebar is always open and cannot be closed
"""


class SidebarOpenSpec(TypedDict):
    desktop: SidebarOpenValue
    mobile: SidebarOpenValue | Literal["always-above"]


@dataclass
class SidebarOpen:
    """
    The initial state of the sidebar at `desktop` and `mobile` screen sizes.
    """

    desktop: SidebarOpenValue = "open"
    """The initial state of the sidebar on desktop screen sizes."""

    mobile: SidebarOpenValue | Literal["always-above"] = "closed"
    """The initial state of the sidebar on mobile screen sizes."""

    _VALUES: tuple[SidebarOpenValue, ...] = field(
        default=("open", "closed", "always"),
        repr=False,
        hash=False,
        compare=False,
    )

    _VALUES_MOBILE: tuple[SidebarOpenValue | Literal["always-above"], ...] = field(
        default=("open", "closed", "always", "always-above"),
        repr=False,
        hash=False,
        compare=False,
    )

    def __post_init__(self):
        if self.desktop not in self._VALUES:
            raise ValueError(f"`desktop` must be one of: {self._VALUES}.")
        if self.mobile not in self._VALUES_MOBILE:
            raise ValueError(f"`mobile` must be one of: {self._VALUES_MOBILE}.")

    def _is_always_open(
        self, on: Literal["desktop", "mobile", "both"] = "both"
    ) -> bool:
        desktop = self.desktop == "always"
        mobile = self.mobile in ("always", "always-above")
        switch = {"desktop": desktop, "mobile": mobile, "both": desktop and mobile}
        return switch[on]

    @classmethod
    def _from_string(cls, open: str) -> SidebarOpen:
        """
        Takes a single string describing the initial sidebar state and returns a
        :class:`~shiny.ui.SidebarOpen` object. In general, the value of `open` becomes
        the initial state for both desktop and mobile screen sizes, unless `open` is
        `"desktop"`, in which case the sidebar starts open on desktop screen, closed on
        mobile.

        Parameters
        ----------
        open
            The initial state of the sidebar. The possible values are:

            * `"desktop"`: the sidebar starts open on desktop screen, closed on mobile
            * `"open"`: the sidebar starts open
            * `"closed"`: the sidebar starts closed
            * `"always"`: the sidebar is always open and cannot be closed

        Returns
        -------
        :
            A :class:`~shiny.ui.SidebarOpen` object.
        """
        bad_value = ValueError(
            f"`open` must be a string matching one of: {SidebarOpen._VALUES}."
        )

        if not isinstance(open, str) or len(open) == 0:
            raise bad_value

        if open == "desktop":
            return cls(desktop="open", mobile="closed")
        elif open in cls._VALUES:
            return cls(desktop=open, mobile=open)
        else:
            raise bad_value

    @classmethod
    def _as_open(
        cls,
        open: SidebarOpenSpec | SidebarOpenValue | Literal["desktop"],
    ) -> SidebarOpen:
        if isinstance(open, dict):
            return cls(**open)

        if isinstance(open, str):
            return cls._from_string(open)

        raise ValueError(
            f"""`open` must be one of {SidebarOpen._VALUES}, """
            + "or a dictionary with keys `desktop` and `mobile` using these values."
        )


@no_example()
class Sidebar:
    """
    A sidebar object

    Class returned from :func:`~shiny.ui.sidebar`. Please do not use this
    class directly. Instead, supply the :func:`~shiny.ui.sidebar` object to
    :func:`~shiny.ui.layout_sidebar`.

    Attributes
    ----------
    children
        A tuple of :class:`~htmltools.Tag` objects that are the contents of the sidebar.
    attrs
        A dictionary of attributes that are supplied to the sidebar contents
        :class:`~htmltools.Tag` container.
    width
        A valid CSS unit used for the width of the sidebar.
    position
        Where the sidebar should appear relative to the main content, one of `"left"` or
        `"right"`.
    id
        The resolved ID. Required if wanting to reactively read (or update) the
        `collapsible` state in a Shiny app.
    title
        A character title to be used as the sidebar title, which will be wrapped in a
        `<div>` element with class `sidebar-title`. You can also provide a custom
        :class:`~htmltools.Tag` for the title element, in which case you'll
        likely want to give this element `class = "sidebar-title"`.
    color
        A dictionary with items `"bg"` for background or `"fg"` for foreground color.
    class_
        CSS classes for the sidebar container element, in addition to the fixed
        `.sidebar` class.
    max_height_mobile
        A CSS length unit (passed through :func:`~shiny.ui.css.as_css_unit`) defining
        the maximum height of the horizontal sidebar when viewed on mobile devices. Only
        applies to always-open sidebars on mobile, where by default the sidebar
        container is placed below the main content container on mobile devices.
    gap
        A CSS length unit defining the vertical `gap` (i.e., spacing) between elements
        provided to `*args`.
    padding
        Padding within the sidebar itself. This can be a numeric vector (which will be
        interpreted as pixels) or a character vector with valid CSS lengths. `padding`
        may be one to four values.

        * If a single value, then that value will be used for all four sides.
        * If two, then the first value will be used for the top and bottom, while
          the second value will be used for left and right.
        * If three values, then the first will be used for top, the second will be left
          and right, and the third will be bottom.
        * If four, then the values will be interpreted as top, right, bottom, and left
          respectively.

    Parameters
    ----------
    children
        A tuple of :class:`~htmltools.Tag` objects that are the contents of the sidebar.
    attrs
        A dictionary of attributes that are supplied to the sidebar contents
        :class:`~htmltools.Tag` container.
    width
        A valid CSS unit used for the width of the sidebar.
    position
        Where the sidebar should appear relative to the main content, one of `"left"` or
        `"right"`.
    open
        The initial state of the sidebar. If a string, the possible values are:

        * `"open"`: the sidebar starts open
        * `"closed"`: the sidebar starts closed
        * `"always"`: the sidebar is always open and cannot be closed

        Alternatively, you can provide a dictionary with keys `"desktop"` and `"mobile"`
        to set different initial states for desktop and mobile. For example, when
        `{"desktop": "open", "mobile": "closed"}` the sidebar is initialized in the
        open state on desktop screens or in the closed state on mobile screens. You can
        also choose to place an always-open sidebar above the main content on mobile
        devices by setting `open={"mobile": "always-above"}`.
    id
        A character string. Required if wanting to reactively read (or update) the
        `collapsible` state in a Shiny app.
    title
        A character title to be used as the sidebar title, which will be wrapped in a
        `<div>` element with class `sidebar-title`. You can also provide a custom
        :class:`~htmltools.Tag` for the title element, in which case you'll
        likely want to give this element `class = "sidebar-title"`.
    bg,fg
        A background or foreground color.
    class_
        CSS classes for the sidebar container element, in addition to the fixed
        `.sidebar` class.
    max_height_mobile
        A CSS length unit (passed through :func:`~shiny.ui.css.as_css_unit`) defining
        the maximum height of the horizontal sidebar when viewed on mobile devices. Only
        applies to always-open sidebars on mobile, where by default the sidebar
        container is placed below the main content container on mobile devices.
    gap
        A CSS length unit defining the vertical `gap` (i.e., spacing) between elements
        provided to `*args`.
    padding
        Padding within the sidebar itself. This can be a numeric vector (which will be
        interpreted as pixels) or a character vector with valid CSS lengths. `padding`
        may be one to four values.

        * If a single value, then that value will be used for all four sides.
        * If two, then the first value will be used for the top and bottom, while
          the second value will be used for left and right.
        * If three values, then the first will be used for top, the second will be left
          and right, and the third will be bottom.
        * If four, then the values will be interpreted as top, right, bottom, and left
          respectively.
    """

    def __init__(
        self,
        *,
        children: list[TagChild],
        attrs: TagAttrs,
        position: Literal["left", "right"] = "left",
        open: Optional[SidebarOpenSpec | SidebarOpenValue | Literal["desktop"]] = None,
        width: CssUnit = 250,
        id: Optional[str] = None,
        title: TagChild | str = None,
        fg: Optional[str] = None,
        bg: Optional[str] = None,
        class_: Optional[str] = None,
        max_height_mobile: Optional[str | float] = None,
        gap: Optional[CssUnit] = None,
        padding: Optional[CssUnit | list[CssUnit]] = None,
    ):
        if isinstance(title, (str, int, float)):
            title = tags.header(str(title), class_="sidebar-title")

        self.id: ResolvedId | None = resolve_id_or_none(id)
        self.title = title
        self.class_ = class_
        self.gap = as_css_unit(gap)
        self.padding = as_css_padding(padding)
        # User-provided initial open state
        self._open: SidebarOpen | None = self._as_open(open)
        # Shiny or consumer-provided default open state, change with `_set_default_open()`
        self._default_open = SidebarOpen(desktop="open", mobile="closed")
        self.position = position
        self.width = as_css_unit(width)
        self._max_height_mobile = max_height_mobile
        self.color = {"fg": fg, "bg": bg}
        self.attrs = attrs
        self.children = children

    def open(
        self,
        value: (
            SidebarOpen
            | SidebarOpenSpec
            | SidebarOpenValue
            | Literal["desktop"]
            | None
            | MISSING_TYPE
        ) = MISSING,
    ) -> SidebarOpen:
        """
        Get or set the initial state of the sidebar. Returns a dataclass with `desktop`
        and `mobile` attributes.
        """
        if not isinstance(value, MISSING_TYPE):
            self._open = self._as_open(value)

        if self._open is None:
            return self._default_open
        else:
            return self._open

    @property
    def max_height_mobile(self) -> Optional[str]:
        max_height_mobile = self._max_height_mobile

        if max_height_mobile is not None and not self.open()._is_always_open("mobile"):
            warnings.warn(
                "The `shiny.ui.sidebar(max_height_mobile=)` argument only applies to "
                + "the sidebar when `open` is 'always' or 'always-above' on mobile, but "
                + f"`open` is `'{self.open().mobile}'`. "
                + "The `max_height_mobile` argument will be ignored.",
                # `stacklevel=2`: Refers to the caller of `.max_height_mobile` property method
                stacklevel=2,
            )
            max_height_mobile = None

        return as_css_unit(max_height_mobile)

    @max_height_mobile.setter
    def max_height_mobile(self, max_height_mobile: str | float | None) -> None:
        self._max_height_mobile = max_height_mobile

    def _as_open(
        self,
        open: Optional[
            SidebarOpen | SidebarOpenSpec | SidebarOpenValue | Literal["desktop"]
        ] = None,
    ) -> SidebarOpen | None:
        if open is None:
            return None

        if isinstance(open, SidebarOpen):
            return open

        return SidebarOpen._as_open(open)

    def _is_always_open(
        self, on: Literal["desktop", "mobile", "both"] = "both"
    ) -> bool:
        return self.open()._is_always_open(on)

    def _get_sidebar_id(self) -> Optional[str]:
        """
        Returns the resolved ID of the sidebar, or `None` if the sidebar is always open.
        When the sidebar is collapsible, but the user hasn't provided an ID, a random ID
        is generated and returned.
        """
        if isinstance(self.id, ResolvedId):
            return self.id

        if self._is_always_open():
            return None

        return private_random_id("bslib_sidebar")

    def _collapse_tag(self, id: str | None) -> Tag:
        """Create the <button> tag for the collapse button."""
        is_expanded = self.open().desktop == "open" or self.open().mobile == "open"

        return tags.button(
            _collapse_icon(),
            class_="collapse-toggle",
            type="button",
            title="Toggle sidebar",
            aria_expanded=(
                ("true" if is_expanded else "false")
                if not self._is_always_open()
                else None
            ),
            aria_controls=id if not self._is_always_open() else None,
        )

    def _sidebar_tag(self, id: str | None) -> Tag:
        """Create the `<aside>` tag for the sidebar."""
        is_hidden_initially = (
            self.open().desktop == "closed" or self.open().mobile == "closed"
        )

        return tags.aside(
            {
                "id": id,
                "class": "sidebar",
                "hidden": "true" if is_hidden_initially else None,
            },
            # If the user provided an id, we make the sidebar an input to report state
            {"class": "bslib-sidebar-input"} if self.id is not None else None,
            div(
                {
                    "class": "sidebar-content bslib-gap-spacing",
                    "style": css(
                        gap=self.gap,
                        padding=self.padding,
                    ),
                },
                self.title,
                *self.children,
                self.attrs,
            ),
            class_=self.class_,
        )

    def tagify(self) -> TagList:
        id = self._get_sidebar_id()
        taglist = TagList(self._sidebar_tag(id), self._collapse_tag(id))
        return taglist.tagify()


@add_example()
def sidebar(
    *args: TagChild | TagAttrs,
    position: Literal["left", "right"] = "left",
    open: Optional[SidebarOpenSpec | SidebarOpenValue | Literal["desktop"]] = None,
    width: CssUnit = 250,
    id: Optional[str] = None,
    title: TagChild | str = None,
    bg: Optional[str] = None,
    fg: Optional[str] = None,
    class_: Optional[str] = None,
    max_height_mobile: Optional[str | float] = None,
    gap: Optional[CssUnit] = None,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    **kwargs: TagAttrValue,
) -> Sidebar:
    # See [this article](https://rstudio.github.io/bslib/articles/sidebars.html)
    #   to learn more.
    # TODO-future; If color contrast is implemented. Docs for `bg` and `fg`:
    #     If only one of either is provided, an
    #     accessible contrasting color is provided for the opposite color, e.g. setting
    #     `bg` chooses an appropriate `fg` color.

    """
    Sidebar element

    Create a collapsing sidebar layout by providing a `sidebar()` object to the
    `sidebar=` argument of:

    * :func:`~shiny.ui.layout_sidebar`
      * Creates a sidebar layout component which can be dropped inside any Shiny UI page method (e.g. :func:`~shiny.ui.page_fillable`) or :func:`~shiny.ui.card` context.
    * :func:`~shiny.ui.navset_bar`, :func:`~shiny.ui.navset_card_tab`, and :func:`~shiny.ui.navset_card_pill`
      * Creates a multi page/tab UI with a singular `sidebar()` (which is
        shown on every page/tab).

    Parameters
    ----------
    *args
        Contents of the sidebar. Or tag attributes that are supplied to the
        resolved :class:`~htmltools.Tag` object.
    width
        A valid CSS unit used for the width of the sidebar.
    position
        Where the sidebar should appear relative to the main content, one of `"left"` or
        `"right"`.
    open
        The initial state of the sidebar. If a string, the possible values are:

        * `"open"`: the sidebar starts open
        * `"closed"`: the sidebar starts closed
        * `"always"`: the sidebar is always open and cannot be closed

        Alternatively, you can provide a dictionary with keys `"desktop"` and `"mobile"`
        to set different initial states for desktop and mobile. For example, when
        `{"desktop": "open", "mobile": "closed"}` the sidebar is initialized in the
        open state on desktop screens or in the closed state on mobile screens.
    id
        A character string. Required if wanting to reactively read (or update) the
        `collapsible` state in a Shiny app.
    title
        A character title to be used as the sidebar title, which will be wrapped in a
        `<div>` element with class `sidebar-title`. You can also provide a custom
        :class:`~htmltools.Tag` for the title element, in which case you'll
        likely want to give this element `class = "sidebar-title"`.
    bg,fg
        A background or foreground color.
    class_
        CSS classes for the sidebar container element, in addition to the fixed
        `.sidebar` class.
    max_height_mobile
        A CSS length unit (passed through :func:`~shiny.ui.css.as_css_unit`) defining
        the maximum height of the horizontal sidebar when viewed on mobile devices. Only
        applies to always-open sidebars on mobile, where by default the sidebar
        container is placed below the main content container on mobile devices.
    gap
        A CSS length unit defining the vertical `gap` (i.e., spacing) between elements
        provided to `*args`.
    padding
        Padding within the sidebar itself. This can be a numeric vector (which will be
        interpreted as pixels) or a character vector with valid CSS lengths. `padding`
        may be one to four values.

        * If a single value, then that value will be used for all four sides.
        * If two, then the first value will be used for the top and bottom, while
          the second value will be used for left and right.
        * If three values, then the first will be used for top, the second will be left
          and right, and the third will be bottom.
        * If four, then the values will be interpreted as top, right, bottom, and left
          respectively.
    **kwargs
        Named attributes are supplied to the sidebar content container.

    Returns
    -------
    :
        A :class:`~shiny.ui.Sidebar` object.

    See Also
    --------
    * :func:`~shiny.ui.layout_sidebar`
    * :func:`~shiny.ui.navset_bar`
    * :func:`~shiny.ui.navset_card_tab`
    * :func:`~shiny.ui.navset_card_pill`
    """
    # TODO-future; validate bg, fg, class_

    # TODO-future; implement
    # if fg is None and bg is not None:
    #     fg = get_color_contrast(bg)
    # if bg is None and fg is not None:
    #     bg = get_color_contrast(fg)

    attrs, children = consolidate_attrs(*args, **kwargs)

    return Sidebar(
        children=children,
        attrs=attrs,
        width=width,
        position=position,
        open=open,
        id=id,
        title=title,
        fg=fg,
        bg=bg,
        class_=class_,
        max_height_mobile=max_height_mobile,
        gap=gap,
        padding=padding,
    )


@add_example()
def layout_sidebar(
    sidebar: Sidebar,
    *args: TagChild | TagAttrs,
    fillable: bool = True,
    fill: bool = True,
    bg: Optional[str] = None,
    fg: Optional[str] = None,
    border: Optional[bool] = None,
    border_radius: Optional[bool] = None,
    border_color: Optional[str] = None,
    gap: Optional[CssUnit] = None,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    height: Optional[CssUnit] = None,
    **kwargs: TagAttrValue,
) -> CardItem:
    """
    Sidebar layout

    Create a sidebar layout component which can be dropped inside any Shiny UI page
    method (e.g. :func:`~shiny.ui.page_fillable`) or
    :func:`~shiny.ui.card` context.

    Parameters
    ----------
    *args
        One argument needs to be of class :class:`~shiny.ui.Sidebar` object created by
        :func:`~shiny.ui.sidebar`. The remaining arguments will contain the contents to
        the main content area. Or tag attributes that are supplied to the resolved
        :class:`~htmltools.Tag` object.
    fillable
        Whether or not the main content area should be wrapped in a fillable container.
        See :func:`~shiny.ui.fill.as_fillable_container` for details.
    fill
        Whether or not the sidebar layout should be wrapped in a fillable container. See
        :func:`~shiny.ui.fill.as_fill_item` for details.
    bg,fg
        A background or foreground color.
    border
        Whether or not to show a border around the sidebar layout.
    border_radius
        Whether or not to round the corners of the sidebar layout.
    border_color
        A border color.
    gap
        A CSS length unit defining the vertical `gap` (i.e., spacing) between elements
        provided to `*args`. This value will only be used if `fillable` is `True`.
    padding
        Padding within the sidebar itself. This can be a numeric vector (which will be
        interpreted as pixels) or a character vector with valid CSS lengths. `padding`
        may be one to four values.

        * If a single value, then that value will be used for all four sides.
        * If two, then the first value will be used for the top and bottom, while
          the second value will be used for left and right.
        * If three values, then the first will be used for top, the second will be left
          and right, and the third will be bottom.
        * If four, then the values will be interpreted as top, right, bottom, and left
          respectively.
    height
        Any valid CSS unit to use for the height.

    Returns
    -------
    :
        A :class:`~htmltools.Tag` object.

    See Also
    --------
    * :func:`~shiny.ui.sidebar`
    """

    sidebar, args = _get_layout_sidebar_sidebar(sidebar, args)

    # TODO-future; implement
    # if fg is None and bg is not None:
    #     fg = get_color_contrast(bg)
    # if bg is None and fg is not None:
    #     bg = get_color_contrast(fg)

    attrs, children = consolidate_attrs(*args, **kwargs)

    main = div(
        {
            "class": f"main{' bslib-gap-spacing' if fillable else ''}",
            "style": css(
                gap=as_css_unit(gap),
                padding=as_css_padding(padding),
            ),
        },
        attrs,
        *children,
    )
    if fillable:
        main = as_fillable_container(main)

    if sidebar.open().mobile == "always-above":
        contents = (sidebar, main)
    else:
        contents = (main, sidebar)

    res = div(
        {"class": "bslib-sidebar-layout bslib-mb-spacing"},
        {"class": "sidebar-right"} if sidebar.position == "right" else None,
        {"class": "sidebar-collapsed"} if sidebar.open().desktop == "closed" else None,
        *contents,
        components_dependencies(),
        _sidebar_init_js(),
        data_bslib_sidebar_init="true",
        data_open_desktop=sidebar.open().desktop,
        data_open_mobile=sidebar.open().mobile,
        data_collapsible_mobile=(
            "false" if sidebar.open()._is_always_open("mobile") else "true"
        ),
        data_collapsible_desktop=(
            "false" if sidebar.open()._is_always_open("desktop") else "true"
        ),
        data_bslib_sidebar_border=trinary(border),
        data_bslib_sidebar_border_radius=trinary(border_radius),
        style=css_no_sub(
            **{
                "--_sidebar-width": sidebar.width,
                "--_sidebar-bg": sidebar.color["bg"],
                "--_sidebar-fg": sidebar.color["fg"],
                "--_main-fg": fg,
                "--_main-bg": bg,
                "--bs-card-border-color": border_color,
                "height": as_css_unit(height),
                "--_mobile-max-height": sidebar.max_height_mobile,
            },
        ),
    )
    if fill:
        res = as_fill_item(res)

    return CardItem(res)


def _get_layout_sidebar_sidebar(
    sidebar: Sidebar,
    args: tuple[TagChild | TagAttrs, ...],
) -> tuple[Sidebar, tuple[TagChild | TagAttrs, ...]]:
    updated_args: list[TagChild | TagAttrs] = []
    original_args = tuple(args)

    if not isinstance(sidebar, Sidebar):
        raise ValueError(
            "`layout_sidebar()` is not being supplied with a `sidebar()` object. "
            "Please supply a `sidebar()` object to `layout_sidebar(sidebar)`."
        )

    # Use `original_args` here so `updated_args` can be safely altered in place
    for arg in original_args:
        if isinstance(arg, Sidebar):
            raise ValueError(
                "`layout_sidebar()` is being supplied with multiple `sidebar()` objects. "
                "Please supply only one `sidebar()` object to `layout_sidebar()`."
            )

        else:
            # Keep the arg!
            updated_args.append(arg)

    return (sidebar, tuple(updated_args))


@add_example()
def update_sidebar(
    id: str,
    *,
    show: Optional[bool] = None,
    session: Optional[Session] = None,
) -> None:
    """
    Update a sidebar's visibility.

    Set a :func:`~shiny.ui.sidebar` state during an active Shiny user session.

    Parameters
    ----------
    id
        The `id` of the :func:`~shiny.ui.sidebar` to toggle.
    show
        The desired visible state of the sidebar, where `True` opens the sidebar and `False` closes the sidebar (if not already in that state).
    session
        A Shiny session object (the default should almost always be used).

    See Also
    --------
    * :func:`~shiny.ui.sidebar`
    * :func:`~shiny.ui.layout_sidebar`
    """
    session = require_active_session(session)

    # method: Literal["toggle", "open", "close"]
    # if open is None or open == "toggle":
    #     method = "toggle"
    # elif open is True or open == "open":
    #     method = "open"
    # elif open is False or open == "closed":
    #     method = "close"
    # else:
    #     if open == "always" or open == "desktop":
    #         raise ValueError(
    #             f"`open = '{open}'` is not supported by `update_sidebar()`"
    #         )
    #     raise ValueError(
    #         "open must be NULL (or 'toggle'), TRUE (or 'open'), or FALSE (or 'closed')"
    #     )
    if show is not None:
        method = "open" if bool(show) else "close"

        def callback() -> None:
            session.send_input_message(id, {"method": method})

        session.on_flush(callback, once=True)


def _collapse_icon() -> TagChild:
    return HTML(
        '<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 16 16" class="bi bi-chevron-left collapse-icon" style="fill:currentColor;" aria-hidden="true" role="img" ><path fill-rule="evenodd" d="M11.354 1.646a.5.5 0 0 1 0 .708L5.707 8l5.647 5.646a.5.5 0 0 1-.708.708l-6-6a.5.5 0 0 1 0-.708l6-6a.5.5 0 0 1 .708 0z"></path></svg>'
    )


def _sidebar_init_js() -> Tag:
    # Note: if we want to avoid inline `<script>` tags in the future for
    # initialization code, we might be able to do so by turning the sidebar layout
    # container into a web component
    return tags.script(
        {"data-bslib-sidebar-init": True},
        "bslib.Sidebar.initCollapsibleAll()",
    )



from __future__ import annotations

from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def click_extended_task_button(
    button: controller.InputTaskButton,
) -> None:
    button.expect_state("ready")
    button.click(timeout=0)
    button.expect_state("busy", timeout=0)
    button.expect_state("ready", timeout=0)


def test_input_action_task_button(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    controller.OutputText(page, "mod1-text_counter").expect_value(
        "Button clicked 0 times"
    )

    # Extended task
    button_task = controller.InputTaskButton(page, "mod1-btn")
    button_task.expect_label_ready("Go")
    button_task.expect_auto_reset(True)
    click_extended_task_button(button_task)

    controller.OutputText(page, "mod1-text_counter").expect_value(
        "Button clicked 1 times"
    )



# from pathlib import Path
from pathlib import PurePath
from typing import Optional

from htmltools import HTMLDependency, Tag

from shiny.module import resolve_id
from shiny.render.renderer import Jsonifiable, Renderer, ValueFn

# This object is used to let Shiny know where the dependencies needed to run
# our component all live. In this case, we're just using a single javascript
# file but we could also include CSS.
custom_component_deps = HTMLDependency(
    "custom-component",
    "1.0.0",
    source={
        "package": "custom_component",
        "subdir": str(PurePath(__file__).parent / "distjs"),
    },
    script={"src": "index.js", "type": "module"},
)


class render_custom_component(Renderer[int]):
    """
    Render a value in a custom component.
    """

    # The UI used within Shiny Express mode
    def auto_output_ui(self) -> Tag:
        return custom_component(self.output_id, height=self.height)

    # The init method is used to set up the renderer's parameters.
    # If no parameters are needed, then the `__init__()` method can be omitted.
    def __init__(self, _fn: Optional[ValueFn[int]] = None, *, height: str = "200px"):
        super().__init__(_fn)
        self.height: str = height

    # Transforms non-`None` values into a `Jsonifiable` object.
    # If you'd like more control on when and how the value is rendered,
    # please use the `async def render(self)` method.
    async def transform(self, value: int) -> Jsonifiable:
        # Send the results to the client. Make sure that this is a serializable
        # object and matches what is expected in the javascript code.
        return {"value": int(value)}


def custom_component(id: str, height: str = "200px") -> Tag:
    """
    A shiny UI output.

    To be paired with `render_custom_component` decorator within the Shiny server.
    """
    return Tag(
        "custom-component",
        custom_component_deps,
        # Use resolve_id so that our component will work in a module
        id=resolve_id(id),
    )



from shiny import reactive
from shiny.express import input, ui

ui.input_slider("receiver", "Receiver:", min=0, max=100, value=50, step=1, width="100%")
ui.p("Change the min and max values below to see the receiver slider above update.")

with ui.layout_column_wrap(width=1 / 2):
    ui.input_slider("min", "Min:", min=0, max=50, value=0, step=1)
    ui.input_slider("max", "Max:", min=50, max=100, value=100, step=1)


@reactive.effect
def _():
    # You can update the value, min, max, and step.
    ui.update_slider(
        "receiver",
        value=max(min(input.receiver(), input.max()), input.min()),
        min=input.min(),
        max=input.max(),
    )



from shared import mtcars

from shiny import reactive
from shiny.express import input, render, ui

data = reactive.value(mtcars.iloc[:, range(4)])

with ui.card():
    ui.input_action_button("btn", "Sort on columns 1↑ and 3↓")

    @render.data_frame
    def df():
        return render.DataGrid(data())


@reactive.effect
@reactive.event(input.btn)
async def _():
    await df.update_sort([{"col": 1, "desc": False}, {"col": 3, "desc": True}])



import textwrap

from htmltools import HTMLDocument, TagList, tags

from shiny import ui
from shiny.ui._input_select import _update_options


def test_panel_title():
    x = HTMLDocument(ui.panel_title("Hello Shiny UI")).render()["html"]
    assert x == textwrap.dedent(
        """\
        <!DOCTYPE html>
        <html>
          <head>
            <meta charset="utf-8"/>
            <script type="application/html-dependencies">headcontent_648f698a6bce952fe1b165140306e01f4da9e164[0.0]</script>
            <title>Hello Shiny UI</title>
          </head>
          <body>
            <h2>Hello Shiny UI</h2>
          </body>
        </html>"""
    )

    title = TagList(
        tags.h1("A title"),
        tags.script("foo"),
        tags.style("foo"),
        tags.h5(tags.script("foo"), "A subtitle"),
    )

    x = HTMLDocument(ui.panel_title(title)).render()["html"]
    assert x == textwrap.dedent(
        """\
        <!DOCTYPE html>
        <html>
          <head>
            <meta charset="utf-8"/>
            <script type="application/html-dependencies">headcontent_9e055af8ef9fa0fb1ba72eeba8053498fbc0d075[0.0]</script>
            <title>A title A subtitle</title>
          </head>
          <body>
            <h1>A title</h1>
            <script>foo</script>
            <style>foo</style>
            <h5>
              <script>foo</script>
              A subtitle
            </h5>
          </body>
        </html>"""
    )


def test_modal_footer():
    # Default behavior: Dismiss button
    x = str(ui.modal())
    assert x == textwrap.dedent(
        """\
        <div id="shiny-modal" class="modal fade" tabindex="-1" data-backdrop="static" data-bs-backdrop="static" data-keyboard="false" data-bs-keyboard="false">
          <div class="modal-dialog">
            <div class="modal-content">
              <div class="modal-body"></div>
              <div class="modal-footer">
                <button class="btn btn-default" type="button" data-dismiss="modal" data-bs-dismiss="modal">Dismiss</button>
              </div>
            </div>
          </div>
          <script>if (window.bootstrap && !window.bootstrap.Modal.VERSION.match(/^4\\. /)) {
          var modal=new bootstrap.Modal(document.getElementById('shiny-modal'))
          modal.show()
        } else {
          $('#shiny-modal').modal().focus()
        }</script>
        </div>"""
    )

    # None: drop footer altogether
    x = str(ui.modal(footer=None))
    assert x == textwrap.dedent(
        """\
        <div id="shiny-modal" class="modal fade" tabindex="-1" data-backdrop="static" data-bs-backdrop="static" data-keyboard="false" data-bs-keyboard="false">
          <div class="modal-dialog">
            <div class="modal-content">
              <div class="modal-body"></div>
            </div>
          </div>
          <script>if (window.bootstrap && !window.bootstrap.Modal.VERSION.match(/^4\\. /)) {
          var modal=new bootstrap.Modal(document.getElementById('shiny-modal'))
          modal.show()
        } else {
          $('#shiny-modal').modal().focus()
        }</script>
        </div>"""
    )

    # If other falsy value: Render empty footer
    x = str(ui.modal(footer=""))
    assert x == textwrap.dedent(
        """\
        <div id="shiny-modal" class="modal fade" tabindex="-1" data-backdrop="static" data-bs-backdrop="static" data-keyboard="false" data-bs-keyboard="false">
          <div class="modal-dialog">
            <div class="modal-content">
              <div class="modal-body"></div>
              <div class="modal-footer"></div>
            </div>
          </div>
          <script>if (window.bootstrap && !window.bootstrap.Modal.VERSION.match(/^4\\. /)) {
          var modal=new bootstrap.Modal(document.getElementById('shiny-modal'))
          modal.show()
        } else {
          $('#shiny-modal').modal().focus()
        }</script>
        </div>"""
    )

    # Anything else: include custom footer
    x = str(ui.modal(footer=ui.span("Custom Footer", class_="mt-3")))
    assert x == textwrap.dedent(
        """\
        <div id="shiny-modal" class="modal fade" tabindex="-1" data-backdrop="static" data-bs-backdrop="static" data-keyboard="false" data-bs-keyboard="false">
          <div class="modal-dialog">
            <div class="modal-content">
              <div class="modal-body"></div>
              <div class="modal-footer">
                <span class="mt-3">Custom Footer</span>
              </div>
            </div>
          </div>
          <script>if (window.bootstrap && !window.bootstrap.Modal.VERSION.match(/^4\\. /)) {
          var modal=new bootstrap.Modal(document.getElementById('shiny-modal'))
          modal.show()
        } else {
          $('#shiny-modal').modal().focus()
        }</script>
        </div>"""
    )


def test__update_options():
    # User does not supply options
    assert _update_options({}, remove_button=False, multiple=True) == {}
    assert _update_options({}, remove_button=True, multiple=True) == {
        "plugins": ["remove_button"]
    }
    assert _update_options({}, remove_button=True, multiple=False) == {
        "plugins": ["clear_button"]
    }

    # User supplies other plugins
    d1 = {"plugins": ["foo", "bar"]}
    assert _update_options(d1, remove_button=True, multiple=True) == {
        "plugins": ["foo", "bar", "remove_button"]
    }
    assert _update_options(d1, remove_button=True, multiple=False) == {
        "plugins": ["foo", "bar", "clear_button"]
    }
    assert _update_options(d1, remove_button=False, multiple=False) == {
        "plugins": ["foo", "bar"]
    }

    # User supplies non-plugin options
    d2 = {"other_key": "foo"}

    assert _update_options(d2, remove_button=True, multiple=True) == {
        "other_key": "foo",
        "plugins": ["remove_button"],
    }
    assert _update_options(d2, remove_button=True, multiple=True) == {
        "other_key": "foo",
        "plugins": ["remove_button"],
    }
    assert _update_options(d2, remove_button=False, multiple=False) == d2

    # User supplies clear button plugin
    d3 = {"plugins": ["clear_button"]}

    assert _update_options(d3, remove_button=True, multiple=True) == {
        "plugins": ["clear_button", "remove_button"]
    }

    assert _update_options(d3, remove_button=False, multiple=True) == d3
    assert _update_options(d3, remove_button=True, multiple=False) == d3

    # User supplies both clear and remove button plugins
    d4 = {"plugins": ["clear_button", "remove_button"]}

    assert _update_options(d4, remove_button=True, multiple=True) == d4
    assert _update_options(d4, remove_button=True, multiple=False) == d4
    assert _update_options(d4, remove_button=False, multiple=False) == d4



from .custom_component import (
    input_custom_component,
    output_custom_component,
    render_custom_component,
)

__all__ = [
    "input_custom_component",
    "output_custom_component",
    "render_custom_component",
]



from shiny import reactive
from shiny.express import module, render, ui


@module
def counter(input, output, session, starting_value: int = 0):
    count = reactive.value(starting_value)

    ui.input_action_button("btn", "Increment")

    with ui.div():

        @render.express
        def current_count():
            count()

    @reactive.effect
    @reactive.event(input.btn)
    def increment():
        count.set(count() + 1)


counter("one")
ui.hr()
counter("two")



from shiny.express import ui

with ui.card(full_screen=True):
    ui.card_header("This is the header")
    ui.p("This is the body.")
    ui.p("This is still the body.")
    ui.card_footer("This is the footer")



import sys

if "pyodide" in sys.modules:
    # psutil doesn't work on pyodide--use fake data instead
    from fakepsutil import cpu_count, cpu_percent
else:
    from psutil import cpu_count, cpu_percent

from math import ceil

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from shiny import App, Inputs, Outputs, Session, reactive, render, ui

# The agg matplotlib backend seems to be a little more efficient than the default when
# running on macOS, and also gives more consistent results across operating systems
matplotlib.use("agg")

# max number of samples to retain
MAX_SAMPLES = 1000
# secs between samples
SAMPLE_PERIOD = 1


ncpu = cpu_count(logical=True)

app_ui = ui.page_fluid(
    ui.tags.style(
        """
        /* Don't apply fade effect, it's constantly recalculating */
        .recalculating, .recalculating > * {
            opacity: 1 !important;
        }
        tbody > tr:last-child {
            /*border: 3px solid var(--bs-dark);*/
            box-shadow:
                0 0 2px 1px #fff, /* inner white */
                0 0 4px 2px #0ff, /* middle cyan */
                0 0 5px 3px #00f; /* outer blue */
        }
        #table table {
            table-layout: fixed;
            width: %s;
            font-size: 0.8em;
        }
        th, td {
            text-align: center;
        }
        """
        % f"{ncpu * 4}em"
    ),
    # Disable busy indicators
    ui.busy_indicators.use(spinners=False, pulse=False),
    ui.h3("CPU Usage %", class_="mt-2"),
    ui.layout_sidebar(
        ui.sidebar(
            ui.input_select(
                "cmap",
                "Colormap",
                {
                    "inferno": "inferno",
                    "viridis": "viridis",
                    "copper": "copper",
                    "prism": "prism (not recommended)",
                },
            ),
            ui.p(ui.input_action_button("reset", "Clear history", class_="btn-sm")),
            ui.input_switch("hold", "Freeze output", value=False),
            class_="mb-3",
        ),
        ui.div(
            {"class": "card mb-3"},
            ui.div(
                {"class": "card-body"},
                ui.h5({"class": "card-title mt-0"}, "Graphs"),
                ui.output_plot("plot", height=f"{ncpu * 40}px"),
            ),
            ui.div(
                {"class": "card-footer"},
                ui.input_numeric("sample_count", "Number of samples per graph", 50),
            ),
        ),
        ui.div(
            {"class": "card"},
            ui.div(
                {"class": "card-body"},
                ui.h5({"class": "card-title m-0"}, "Heatmap"),
            ),
            ui.div(
                {"class": "card-body overflow-auto pt-0"},
                ui.output_table("table"),
            ),
            ui.div(
                {"class": "card-footer"},
                ui.input_numeric("table_rows", "Rows to display", 5),
            ),
        ),
    ),
)


@reactive.calc
def cpu_current():
    reactive.invalidate_later(SAMPLE_PERIOD)
    return cpu_percent(percpu=True)


def server(input: Inputs, output: Outputs, session: Session):
    cpu_history = reactive.value(None)

    @reactive.calc
    def cpu_history_with_hold():
        # If "hold" is on, grab an isolated snapshot of cpu_history; if not, then do a
        # regular read
        if not input.hold():
            return cpu_history()
        else:
            # Even if frozen, we still want to respond to input.reset()
            input.reset()
            with reactive.isolate():
                return cpu_history()

    @reactive.effect
    def collect_cpu_samples():
        """cpu_percent() reports just the current CPU usage sample; this Effect gathers
        them up and stores them in the cpu_history reactive value, in a numpy 2D array
        (rows are CPUs, columns are time)."""

        new_data = np.vstack(cpu_current())
        with reactive.isolate():
            if cpu_history() is None:
                cpu_history.set(new_data)
            else:
                combined_data = np.hstack([cpu_history(), new_data])
                # Throw away extra data so we don't consume unbounded amounts of memory
                if combined_data.shape[1] > MAX_SAMPLES:
                    combined_data = combined_data[:, -MAX_SAMPLES:]
                cpu_history.set(combined_data)

    @reactive.effect(priority=100)
    @reactive.event(input.reset)
    def reset_history():
        cpu_history.set(None)

    @render.plot
    def plot():
        history = cpu_history_with_hold()

        if history is None:
            history = np.array([])
            history.shape = (ncpu, 0)

        nsamples = input.sample_count()

        # Throw away samples too old to fit on the plot
        if history.shape[1] > nsamples:
            history = history[:, -nsamples:]

        ncols = 2
        nrows = int(ceil(ncpu / ncols))
        fig, axeses = plt.subplots(
            nrows=nrows,
            ncols=ncols,
            squeeze=False,
        )
        for i in range(0, ncols * nrows):
            row = i // ncols
            col = i % ncols
            axes = axeses[row, col]
            if i >= len(history):
                axes.set_visible(False)
                continue
            data = history[i]
            axes.yaxis.set_label_position("right")
            axes.yaxis.tick_right()
            axes.set_xlim(-(nsamples - 1), 0)
            axes.set_ylim(0, 100)

            assert len(data) <= nsamples

            # Set up an array of x-values that will right-align the data relative to the
            # plotting area
            x = np.arange(0, len(data))
            x = np.flip(-x)

            # Color bars by cmap
            color = plt.get_cmap(input.cmap())(data / 100)
            axes.bar(x, data, color=color, linewidth=0, width=1.0)

            axes.set_yticks([25, 50, 75])
            for ytl in axes.get_yticklabels():
                if col == ncols - 1 or i == ncpu - 1 or True:
                    ytl.set_fontsize(7)
                else:
                    ytl.set_visible(False)
                    hide_ticks(axes.yaxis)
            for xtl in axes.get_xticklabels():
                xtl.set_visible(False)
            hide_ticks(axes.xaxis)
            axes.grid(True, linewidth=0.25)

        return fig

    @render.table
    def table():
        history = cpu_history_with_hold()
        latest = pd.DataFrame(history).transpose().tail(input.table_rows())
        if latest.shape[0] == 0:
            return latest
        return (
            latest.style.format(precision=0)
            .hide(axis="index")
            .set_table_attributes(
                'class="dataframe shiny-table table table-borderless font-monospace"'
            )
            .background_gradient(cmap=input.cmap(), vmin=0, vmax=100)
        )


def hide_ticks(axis):
    for ticks in [axis.get_major_ticks(), axis.get_minor_ticks()]:
        for tick in ticks:
            tick.tick1line.set_visible(False)
            tick.tick2line.set_visible(False)
            tick.label1.set_visible(False)
            tick.label2.set_visible(False)


app = App(app_ui, server)



from __future__ import annotations

import re

from playwright.sync_api import Locator, Page
from playwright.sync_api import expect as playwright_expect

from .._types import PatternOrStr, Timeout
from ..expect._internal import (
    expect_attribute_to_have_value as _expect_attribute_to_have_value,
)
from ._base import InitLocator, UiBase


class _OverlayBase(UiBase):
    """Base class for overlay controls"""

    loc_trigger: Locator
    """
    Playwright `Locator` for the trigger element.
    """
    loc_overlay_body: Locator
    """
    Playwright `Locator` for the overlay body.
    """
    loc_overlay_container: Locator
    """
    Playwright `Locator` for of the overlay container.
    """

    def __init__(
        self,
        page: Page,
        *,
        id: str,
        loc: InitLocator,
        overlay_name: str,
        overlay_selector: str,
    ) -> None:
        """
        Initializes a new instance of the `OverlayBase` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the overlay.
        loc
            Playwright `Locator` of the overlay.
        overlay_name
            The name of the overlay.
        overlay_selector
            The selector of the overlay.
        """
        super().__init__(page, id=id, loc=loc)
        self._overlay_name = overlay_name
        self._overlay_selector = overlay_selector
        self.loc_trigger = self.loc.locator(
            f" > :last-child[data-bs-toggle='{self._overlay_name}']"
        )

    def _get_overlay_id(self, *, timeout: Timeout = None) -> str | None:
        """Note. This requires 2 steps. Will not work if the overlay element is rapidly created during locator fetch"""
        loc_el = self.loc.locator(
            f" > :last-child[data-bs-toggle='{self._overlay_name}']"
        )
        loc_el.wait_for(state="visible", timeout=timeout)
        loc_el.scroll_into_view_if_needed(timeout=timeout)
        return loc_el.get_attribute("aria-describedby")

    # @property
    # def loc_overlay_body(self) -> Locator:
    #     # Can not leverage `self.loc_overlay_container` as `self._overlay_selector` must
    #     # be concatenated directly to the result of `self._get_overlay_id()`
    #     return self.page.locator(f"#{self._get_overlay_id()}{self._overlay_selector}")

    # @property
    # def loc_overlay_container(self) -> Locator:
    #     return self.page.locator(f"#{self._get_overlay_id()}")

    def get_loc_overlay_body(self, *, timeout: Timeout = None) -> Locator:
        # Can not leverage `self.loc_overlay_container` as `self._overlay_selector` must
        # be concatenated directly to the result of `self._get_overlay_id()`
        return self.page.locator(
            f"#{self._get_overlay_id(timeout=timeout)}{self._overlay_selector}"
        )

    def get_loc_overlay_container(self, *, timeout: Timeout = None) -> Locator:
        """
        Returns the locator for the overlay container.

        Parameters
        ----------
        timeout
            The maximum time to wait for the overlay container to appear. Defaults to `None`.
        """
        return self.page.locator(f"#{self._get_overlay_id(timeout=timeout)}")

    def expect_body(self, value: PatternOrStr, *, timeout: Timeout = None) -> None:
        """
        Expects the overlay body to have the specified text.

        Parameters
        ----------
        value
            The expected text pattern or string.
        timeout
            The maximum time to wait for the overlay body to appear. Defaults to `None`.
        """
        playwright_expect(self.get_loc_overlay_body(timeout=timeout)).to_have_text(
            value, timeout=timeout
        )

    def expect_active(self, value: bool, *, timeout: Timeout = None) -> None:
        """
        Expects the overlay to be active or inactive.

        Parameters
        ----------
        value
            `True` if the overlay is expected to be active, False otherwise.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        attr_value = re.compile(r".*") if value else None
        return _expect_attribute_to_have_value(
            loc=self.loc_trigger,
            timeout=timeout,
            name="aria-describedby",
            value=attr_value,
        )

    def expect_placement(self, value: str, *, timeout: Timeout = None) -> None:
        """
        Expects the overlay to have the specified placement.

        Parameters
        ----------
        value
            The expected placement value.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        return _expect_attribute_to_have_value(
            loc=self.get_loc_overlay_container(timeout=timeout),
            timeout=timeout,
            name="data-popper-placement",
            value=value,
        )


class Popover(_OverlayBase):
    """Controller for :func:`shiny.ui.popover`."""

    loc_trigger: Locator
    """
    Playwright `Locator` for the trigger element that opens/closes the popover.
    """
    loc_overlay_body: Locator
    """
    Playwright `Locator` for the popover body.
    """
    loc_overlay_container: Locator
    """
    Playwright `Locator` for the popover container.
    """

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a new instance of the `Popover` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the popover.
        """
        super().__init__(
            page,
            id=id,
            loc=f"bslib-popover#{id}",
            overlay_name="popover",
            overlay_selector=".popover > div.popover-body",
        )

    def set(self, open: bool, timeout: Timeout = None) -> None:
        """
        Sets the state of the popover.

        Parameters
        ----------
        open
            `True` to open the popover and `False` to close it.
        timeout
            The maximum time to wait for the popover to be visible and interactable. Defaults to `None`.
        """
        if open ^ self.get_loc_overlay_body(timeout=timeout).count() > 0:
            self._toggle(timeout=timeout)

    def _toggle(self, timeout: Timeout = None) -> None:
        """
        Toggles the state of the popover.

        Parameters
        ----------
        timeout
            The maximum time to wait for the popover to be visible and interactable. Defaults to `None`.
        """
        self.loc_trigger.wait_for(state="visible", timeout=timeout)
        self.loc_trigger.scroll_into_view_if_needed(timeout=timeout)
        self.loc_trigger.click(timeout=timeout)

    def expect_title(self, value: PatternOrStr, *, timeout: Timeout = None) -> None:
        """
        Expects the popover title to have the specified text.

        Parameters
        ----------
        value
            The expected text pattern or string.
        timeout
            The maximum time to wait for the popover header to appear. Defaults to `None`.
        """
        playwright_expect(
            self.get_loc_overlay_container().locator("> .popover-header")
        ).to_have_text(value, timeout=timeout)


class Tooltip(_OverlayBase):
    """Controller for :func:`shiny.ui.tooltip`."""

    loc_container: Locator
    """
    Playwright `Locator` for the container tooltip.
    """
    loc: Locator
    """
    Playwright `Locator` for the tooltip content.
    """
    loc_trigger: Locator
    """
    Playwright `Locator` for the trigger element.
    """
    loc_overlay_body: Locator
    """
    Playwright `Locator` for the overlay body.
    """
    loc_overlay_container: Locator
    """
    Playwright `Locator` for the overlay container.
    """

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a new instance of the `Tooltip` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the tooltip.
        """
        super().__init__(
            page,
            id=id,
            loc=f"bslib-tooltip#{id}",
            overlay_name="tooltip",
            overlay_selector=".tooltip > div.tooltip-inner",
        )

    def set(self, open: bool, timeout: Timeout = None) -> None:
        """
        Sets the state of the tooltip.

        Parameters
        ----------
        open
            `True` to open the tooltip and `False` to close it.
        timeout
            The maximum time to wait for the tooltip to be visible and interactable. Defaults to `None`.
        """
        if open ^ self.get_loc_overlay_body(timeout=timeout).count() > 0:
            self._toggle(timeout=timeout)
        if not open:
            self.get_loc_overlay_body(timeout=timeout).click()

    def _toggle(self, timeout: Timeout = None) -> None:
        """
        Toggles the state of the tooltip.

        Parameters
        ----------
        timeout
            The maximum time to wait for the tooltip to be visible and interactable. Defaults to `None`.
        """
        self.loc_trigger.wait_for(state="visible", timeout=timeout)
        self.loc_trigger.scroll_into_view_if_needed(timeout=timeout)
        self.loc_trigger.hover(timeout=timeout)



from shiny import reactive
from shiny.express import input, ui

ui.input_slider("controller", "Controller", min=0, max=20, value=10)
ui.input_numeric("inNumber", "Input number", 0)
ui.input_numeric("inNumber2", "Input number 2", 0)


@reactive.effect
def _():
    x = input.controller()
    ui.update_numeric("inNumber", value=x)
    ui.update_numeric(
        "inNumber2",
        label="Number label " + str(x),
        value=x,
        min=x - 10,
        max=x + 10,
        step=5,
    )



from shiny import reactive
from shiny.express import render, ui

ui.page_opts(title="Hello Chat")

# Create and display the chat
chat = ui.Chat(id="chat")
chat.ui()


@reactive.effect
async def _():
    await chat.append_message_stream(("FIRST ", "FIRST ", "FIRST"))


@reactive.effect
async def _():
    await chat.append_message("SECOND SECOND SECOND")


@reactive.effect
async def _():
    await chat.append_message_stream(("THIRD ", "THIRD ", "THIRD"))


@reactive.effect
async def _():
    await chat.append_message("FOURTH FOURTH FOURTH")


@reactive.effect
async def _():
    await chat.append_message_stream(("FIFTH ", "FIFTH ", "FIFTH"))


"Message state:"


@render.code
def message_state():
    return str(chat.messages())



import asyncio
import datetime
import sqlite3
from pathlib import Path

import numpy as np
import pandas as pd

here = Path(__file__).parent
accuracy_scores = pd.read_csv(here / "fake_accuracy_scores.csv")
accuracy_scores.set_index("second", inplace=True)
(here / "data").mkdir(exist_ok=True)

SQLITE_DB_URI = f"file:{here / 'data' / 'accuracy_scores.sqlite'}"


def init_db():
    with sqlite3.connect(SQLITE_DB_URI, uri=True, timeout=30) as con:
        con.execute("PRAGMA journal_mode=WAL")
        con.execute("drop table if exists accuracy_scores")

        now = datetime.datetime.utcnow()
        position = now.minute * 60 + now.second + 1

        # Simulate 100 seconds of historical data
        offset_secs = -np.arange(100) - 1
        abs_secs = (position + offset_secs) % (60 * 60) + 1
        initial_scores = accuracy_scores.loc[abs_secs]
        timestamps = pd.DataFrame(
            {
                "timestamp": now + pd.to_timedelta(offset_secs, unit="s"),
                "second": abs_secs,
            }
        ).set_index("second", inplace=False)
        initial_scores = initial_scores.join(timestamps, how="left")
        initial_scores.to_sql("accuracy_scores", con, index=False, if_exists="append")

        con.execute(
            "create index idx_accuracy_scores_timestamp on accuracy_scores(timestamp)"
        )

        return position


async def update_db(position):
    with sqlite3.connect(SQLITE_DB_URI, uri=True, timeout=30) as con:
        while True:
            new_data = accuracy_scores.loc[position].copy()
            # del new_data["second"]
            new_data["timestamp"] = datetime.datetime.utcnow()
            new_data.to_sql("accuracy_scores", con, index=False, if_exists="append")
            position = (position % (60 * 60)) + 1
            await asyncio.sleep(1)


def begin():
    position = init_db()

    # After initializing the database, we need to start a non-blocking task to update it
    # every second or so. If an event loop is already running, we can use an asyncio
    # task. (This is the case when running via `shiny run` and shinylive.) Otherwise, we
    # need to launch a background thread and run an asyncio event loop there. (This is
    # the case when running via shinyapps.io or Posit Connect.)

    if asyncio.get_event_loop().is_running():
        asyncio.create_task(update_db(position))
    else:
        from threading import Thread

        Thread(target=lambda: asyncio.run(update_db(position)), daemon=True).start()



from shiny import App, Inputs, Outputs, Session, reactive, req, ui

app_ui = ui.page_fluid(
    ui.input_action_button("btn_show", "Show popover", class_="mt-3 me-3"),
    ui.input_action_button("btn_close", "Close popover", class_="mt-3 me-3"),
    ui.br(),
    ui.br(),
    ui.popover(
        ui.input_action_button("btn_w_popover", "A button w/ a popover", class_="mt-3"),
        "A message",
        id="popover_id",
        placement="bottom",
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.effect
    def _():
        req(input.btn_show())

        ui.update_popover("popover_id", show=True)

    @reactive.effect
    def _():
        req(input.btn_close())

        ui.update_popover("popover_id", show=False)

    @reactive.effect
    def _():
        req(input.btn_w_popover())
        ui.notification_show("Button clicked!", duration=3, type="message")


app = App(app_ui, server=server)



from shiny.express import ui

ui.page_opts(
    id="default_page_navbar",
    title="Default Page Navbar",
    window_title="Page NavBar title",
    lang="en",
    fillable_mobile=True,
    fillable=True,
)

with ui.nav_panel("Data"):
    "This page could be used to pick a dataset."

with ui.nav_panel("View"):
    "This page could be used to view the dataset."



import sys
from typing import TYPE_CHECKING, Literal, Union

from ._chat_types import ChatMessage

if TYPE_CHECKING:
    from anthropic.types import MessageParam as AnthropicMessage
    from langchain_core.messages import AIMessage, HumanMessage, SystemMessage
    from ollama import Message as OllamaMessage
    from openai.types.chat import (
        ChatCompletionAssistantMessageParam,
        ChatCompletionSystemMessageParam,
        ChatCompletionUserMessageParam,
    )

    if sys.version_info >= (3, 9):
        import google.generativeai.types as gtypes  # pyright: ignore[reportMissingTypeStubs]

        GoogleMessage = gtypes.ContentDict
    else:
        GoogleMessage = object

    LangChainMessage = Union[AIMessage, HumanMessage, SystemMessage]
    OpenAIMessage = Union[
        ChatCompletionAssistantMessageParam,
        ChatCompletionSystemMessageParam,
        ChatCompletionUserMessageParam,
    ]

    ProviderMessage = Union[
        AnthropicMessage, GoogleMessage, LangChainMessage, OpenAIMessage, OllamaMessage
    ]
else:
    AnthropicMessage = GoogleMessage = LangChainMessage = OpenAIMessage = (
        OllamaMessage
    ) = ProviderMessage = object

ProviderMessageFormat = Literal[
    "anthropic",
    "google",
    "langchain",
    "openai",
    "ollama",
]


# TODO: use a strategy pattern to allow others to register
# their own message formats
def as_provider_message(
    message: ChatMessage, format: ProviderMessageFormat
) -> "ProviderMessage":
    if format == "anthropic":
        return as_anthropic_message(message)
    if format == "google":
        return as_google_message(message)
    if format == "langchain":
        return as_langchain_message(message)
    if format == "openai":
        return as_openai_message(message)
    if format == "ollama":
        return as_ollama_message(message)
    raise ValueError(f"Unknown format: {format}")


def as_anthropic_message(message: ChatMessage) -> "AnthropicMessage":
    from anthropic.types import MessageParam as AnthropicMessage

    if message["role"] == "system":
        raise ValueError(
            "Anthropic requires a system prompt to be specified in the `.create()` method"
        )
    return AnthropicMessage(content=message["content"], role=message["role"])


def as_google_message(message: ChatMessage) -> "GoogleMessage":
    if sys.version_info < (3, 9):
        raise ValueError("Google requires Python 3.9")

    import google.generativeai.types as gtypes  # pyright: ignore[reportMissingTypeStubs]

    role = message["role"]

    if role == "system":
        raise ValueError(
            "Google requires a system prompt to be specified in the `GenerativeModel()` constructor."
        )
    elif role == "assistant":
        role = "model"
    return gtypes.ContentDict(parts=[message["content"]], role=role)


def as_langchain_message(message: ChatMessage) -> "LangChainMessage":
    from langchain_core.messages import AIMessage, HumanMessage, SystemMessage

    content = message["content"]
    role = message["role"]
    if role == "system":
        return SystemMessage(content=content)
    if role == "assistant":
        return AIMessage(content=content)
    if role == "user":
        return HumanMessage(content=content)
    raise ValueError(f"Unknown role: {message['role']}")


def as_openai_message(message: ChatMessage) -> "OpenAIMessage":
    from openai.types.chat import (
        ChatCompletionAssistantMessageParam,
        ChatCompletionSystemMessageParam,
        ChatCompletionUserMessageParam,
    )

    content = message["content"]
    role = message["role"]
    if role == "system":
        return ChatCompletionSystemMessageParam(content=content, role=role)
    if role == "assistant":
        return ChatCompletionAssistantMessageParam(content=content, role=role)
    if role == "user":
        return ChatCompletionUserMessageParam(content=content, role=role)
    raise ValueError(f"Unknown role: {role}")


def as_ollama_message(message: ChatMessage) -> "OllamaMessage":
    from ollama import Message as OllamaMessage

    return OllamaMessage(content=message["content"], role=message["role"])



# -----------------------------------------------------------------------------
# An example of placing a Shiny Chat instance in a sidebar (and having it fill the sidebar).
# To run it, you'll need an OpenAI API key.
# To get one, follow the instructions at https://platform.openai.com/docs/quickstart
# -----------------------------------------------------------------------------
import os

from langchain_openai import ChatOpenAI

from shiny.express import ui

# Provide your API key here (or set the environment variable)
llm = ChatOpenAI(
    api_key=os.environ.get("OPENAI_API_KEY"),  # type: ignore
)

# Set some Shiny page options
ui.page_opts(
    title="Hello Sidebar Chat",
    fillable=True,
    fillable_mobile=True,
)

# Create a chat instance, with an initial message
chat = ui.Chat(
    id="chat",
    messages=[
        {"content": "Hello! How can I help you today?", "role": "assistant"},
    ],
)

# Display the chat in a sidebar
with ui.sidebar(width=300, style="height:100%", position="right"):
    chat.ui(height="100%")


# Define a callback to run when the user submits a message
@chat.on_user_submit
async def _():
    # Get messages currently in the chat
    messages = chat.messages(format="langchain")
    # Create a response message stream
    response = llm.astream(messages)
    # Append the response stream into the chat
    await chat.append_message_stream(response)


"Lorem ipsum dolor sit amet, consectetur adipiscing elit"



# This file should be copyable to the rsconnect-python package and usable without
# modification. The is_express_app() function should not be moved out of this file, and
# this file should not import anything else from shiny.

from __future__ import annotations

import ast
import re
import sys
from pathlib import Path
from typing import Literal, cast

from .._docstring import no_example

__all__ = ("is_express_app",)


@no_example()
def is_express_app(app: str, app_dir: str | None) -> bool:
    """Detect whether an app file is a Shiny express app

    Parameters
    ----------
    app
        App filename, like "app.py". It may be a relative path or absolute path.
    app_dir
        Directory containing the app file. If this is `None`, then `app` must be an
        absolute path.

    Returns
    -------
    :
        `True` if it is a Shiny express app, `False` otherwise.
    """
    if not app.lower().endswith(".py"):
        return False

    if app_dir is not None:
        app_path = Path(app_dir) / app
    else:
        app_path = Path(app)

    if not app_path.exists():
        return False

    try:
        # Read the file, parse it, and look for any imports of shiny.express.
        with open(app_path, encoding="utf-8") as f:
            content = f.read()

        # Check for magic comment in the first 1000 characters
        forced_mode = find_magic_comment_mode(content[:1000])
        if forced_mode == "express":
            return True
        elif forced_mode == "core":
            return False

        tree = ast.parse(content, app_path)
        detector = DetectShinyExpressVisitor()
        detector.visit(tree)

    except Exception:
        return False

    return detector.found_shiny_express_import


class DetectShinyExpressVisitor(ast.NodeVisitor):
    def __init__(self):
        super().__init__()
        self.found_shiny_express_import = False

    def visit_Import(self, node: ast.Import) -> None:
        if any(alias.name == "shiny.express" for alias in node.names):
            self.found_shiny_express_import = True

    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
        if node.module == "shiny.express":
            self.found_shiny_express_import = True
        elif node.module == "shiny" and any(
            alias.name == "express" for alias in node.names
        ):
            self.found_shiny_express_import = True

    # Visit top-level nodes.
    def visit_Module(self, node: ast.Module) -> None:
        super().generic_visit(node)

    # Don't recurse into any nodes, so the we'll only ever look at top-level nodes.
    def generic_visit(self, node: ast.AST) -> None:
        pass


def find_magic_comment_mode(content: str) -> Literal["core", "express"] | None:
    """
    Look for a magic comment of the form "# shiny_mode: express" or "# shiny_mode:
    core".

    If a line of the form "# shiny_mode: x" is found, where "x" is not "express" or
    "core", then a message will be printed to stderr.

    Returns
    -------
    :
        `"express"` if Shiny Express comment is found, `"core"` if Shiny Core comment is
        found, and `None` if no magic comment is found.
    """
    m = re.search(r"^#[ \t]*shiny_mode:[ \t]*(\S*)[ \t]*$", content, re.MULTILINE)
    if m is not None:
        shiny_mode = cast(str, m.group(1))
        if shiny_mode in ("express", "core"):
            # The "type: ignore" is needed for mypy, which is used on some projects that
            # use duplicates of this code.
            return shiny_mode  # type: ignore
        else:
            print(f'Invalid shiny_mode: "{shiny_mode}"', file=sys.stderr)

    return None



from __future__ import annotations

import warnings

# TODO-barret-render.data_frame; Docs
# TODO-barret-render.data_frame; Add examples!
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Literal, Union, cast

from htmltools import Tag

from .. import reactive, ui
from .._docstring import add_example
from .._utils import wrap_async
from .._validation import req
from ..session._utils import require_active_session, session_context
from ..types import JsonifiableDict, ListOrTuple
from ._data_frame_utils._datagridtable import DataGrid, DataTable
from ._data_frame_utils._html import maybe_as_cell_html
from ._data_frame_utils._patch import (
    CellPatch,
    CellValue,
    PatchesFn,
    PatchesFnSync,
    PatchFn,
    PatchFnSync,
    assert_patches_shape,
)
from ._data_frame_utils._reactive_method import reactive_calc_method
from ._data_frame_utils._selection import (
    BrowserCellSelection,
    CellSelection,
    SelectionModes,
    as_cell_selection,
)
from ._data_frame_utils._styles import as_browser_style_infos
from ._data_frame_utils._tbl_data import (
    apply_frame_patches,
    as_data_frame,
    assert_data_is_not_none,
    data_frame_to_native,
    serialize_frame,
    subset_frame,
)
from ._data_frame_utils._types import (
    CellPatchProcessed,
    ColumnFilter,
    ColumnSort,
    DataFrame,
    FrameRender,
    IntoDataFrameT,
    cell_patch_processed_to_jsonifiable,
    frame_render_to_jsonifiable,
)

# as_selection_location_js,
from .renderer import Jsonifiable, Renderer, ValueFn

if TYPE_CHECKING:
    from ..session import Session


@add_example()
class data_frame(
    Renderer[Union[IntoDataFrameT, DataGrid[IntoDataFrameT], DataTable[IntoDataFrameT]]]
):
    """
    Decorator for a function that returns a [pandas](https://pandas.pydata.org/),
    [polars](https://pola.rs/), or eager
    [`narwhals`](https://narwhals-dev.github.io/narwhals/) compatible `DataFrame` object
    to render as an interactive table or grid. Features fast virtualized scrolling,
    sorting, filtering, and row selection (single or multiple).

    Returns
    -------
    :
        A decorator for a function that returns any of the following:

        1. A :class:`~shiny.render.DataGrid` or :class:`~shiny.render.DataTable` object,
           which can be used to customize the appearance and behavior of the data frame
           output.
        2. A [pandas](https://pandas.pydata.org/), [polars](https://pola.rs/), or eager
           [`narwhals`](https://narwhals-dev.github.io/narwhals/) compatible `DataFrame`
           object. This object will be internally upgraded to a default
           `shiny.render.DataGrid(df)`.

    Row selection
    -------------
    When using the row selection feature, you can access the selected rows by using the
    `<data_frame_renderer>.cell_selection()` method, where `<data_frame_renderer>`
    is the `@render.data_frame` function name that corresponds with the `id=` used in
    :func:`~shiny.ui.outout_data_frame`. Internally,
    `<data_frame_renderer>.cell_selection()` retrieves the selected cell
    information from session's `input.<data_frame_renderer>_cell_selection()` value and
    upgrades it for consistent subsetting.

    For example, to filter your pandas data frame (`df`) down to the selected rows you can use:

    * `df.iloc[list(input.<data_frame_renderer>_cell_selection()["rows"])]`
    * `df.iloc[list(<data_frame_renderer>.cell_selection()["rows"])]`
    * `<data_frame_renderer>.data_view(selected=True)`

    The last method (`.data_view(selected=True)`) will also apply any sorting,
    filtering, or edits that has been applied by the user.

    Editing cells
    -------------
    When a returned `DataTable` or `DataGrid` object has `editable=True`, app users will
    be able to edit the cells in the table. After a cell has been edited, the edited
    value will be sent to the server for processing. The handling methods are set via
    `@<data_frame_renderer>.set_patch_fn` or `@<data_frame_renderer>.set_patches_fn`
    decorators. By default, both decorators will return the corresponding value as a
    string.

    Data methods
    ------------

    There are several methods available to inspect and update data frame renderer. It is
    important to know the side effects of each method to know how they interact with
    each other.

    * Data frame render method:
        * When this method is reactively executed, the `.data()` data frame is set to
          the underlying data frame and all `.cell_patches()` are removed.
        * When this method is reactively executed, **all user state is reset**. This
          includes the user's edits, sorting, filtering.
    * `.data()`:
        * Reactive calculation that returns the render method's underlying data frame or
          the data frame supplied to `.update_data(data)`, whichever has been most
          recently set.
    * `.cell_patches()`:
        * Reactive calculation that returns a list of user edits (or updated cell
          values) that will be applied to `.data()` to create the `.data_patched()` data
          frame.
    * `.data_patched()`:
        * Reactive calculation that returns the `.data()` data frame with all
          `.cell_patches()` patches applied.
    * `.data_view(*, selected: bool)`:
        * Reactive function that returns the `.data_patched()` data frame with the
          user's sorting and filtering applied. It represents the data frame as viewed
          by the user within the browser.
        * If `selected=True`, only the selected rows are returned.
    * `.update_cell_value(value, row, col)`:
        * Sets a new entry in `.cell_patches()`.
        * Calling this method will **not** reset the user's sorting or filtering.
    * `.update_data(data)`:
        * Updates the `.data()` data frame with new data.
        * Calling this method will remove all `.cell_patches()`.
        * Calling this method will **not** reset the user's sorting or filtering.

    Note: All data methods are shallow copies of each other. If they are mutated in
    place, it **will modify** the underlying data object and possibly alter other data
    objects.

    Narwhals
    -------------------

    Shiny uses [`narwhals`](https://narwhals-dev.github.io/narwhals/) to manage data
    frame interactions. From their website: "Extremely lightweight and extensible
    compatibility layer between dataframe libraries!". This allows for seamless
    integration between pandas, polars, and any other eagerly defined data frame type.

    There are some reasonable limitations to the narwhals compatibility layer. As they
    are found, they will be added to this list:
    * When converting the column type who does not have a 1:1 mapping between libraries
      (such as pandas' columns containing `str` and `dict` items both share the same
      `object` data type), narwhals will only inspect the first row to disambiguate the
      cell type. This could lead to false negatives in the data type conversion. Shiny
      could inspect each column in an attempt to disambiguate the cell type, but this
      would be a costly operation. The best way to avoid this is to use consistent
      typing. For example, if your first row of the pandas column contains a string and
      the second row of the same column contains a `ui.TagList`, the column will
      incorrectly be interpreted as a string. To get around this, you can wrap all cells
      (or at the very lest the first cell) in the same column within a `ui.TagList` as
      it will not insert any tags, but it will cause the column to be interpreted as
      `html` where possible.   (tl/dr: Use consistent typing in your columns!)

    Tip
    ---
    This decorator should be applied **before** the ``@output`` decorator (if that
    decorator is used). Also, the name of the decorated function (or
    ``@output(id=...)``) should match the ``id`` of a :func:`~shiny.ui.output_data_frame`
    container (see :func:`~shiny.ui.output_data_frame` for example usage).

    See Also
    --------
    * :func:`~shiny.ui.output_data_frame`
    * :class:`~shiny.render.DataGrid` and :class:`~shiny.render.DataTable` are the
      objects you can return from the rendering function to specify options.
    """

    _value: reactive.Value[None | DataGrid[IntoDataFrameT] | DataTable[IntoDataFrameT]]
    """
    Reactive value of the data frame's rendered object.
    """
    _updated_data: reactive.Value[IntoDataFrameT]
    """
    Reactive value of the data frame's updated data value object.
    """

    def _req_value(self):
        """
        Ensure that the value is not `None`.

        Raises a silent error if the value is `None`.

        Returns
        -------
        :
            The non-`None` value of the data frame's rendered object.
        """
        value = self._value()
        if not value:
            req(False)
            raise RuntimeError("req() failed to raise a silent error.")
        return value

    @reactive_calc_method
    def data(self) -> IntoDataFrameT:
        """
        Reactive calculation of the data frame's data.

        This reactive calculation returns the render method's underlying data frame or
        the data frame supplied to `.update_data(data)`, whichever has been most
        recently set.

        The returned value is a shallow copy of the original data frame. It is possible
        that alterations to the `.data()` data frame could alter other associated data
        frame values. Please be cautious when using this value directly.

        Returns
        -------
        :
            This reactive calculation returns the render method's underlying data frame
            or the data frame supplied to `.update_data(data)`, whichever has been most
            recently set.
        """
        # iff updated data exists, return it
        # Can be unset on a followup render call
        if self._updated_data.is_set():
            return self._updated_data()

        return self._req_value().data

    @reactive_calc_method
    def _nw_data(self) -> DataFrame[IntoDataFrameT]:
        """
        Reactive calculation of the data frame's data wrapped by narwhals.

        This is a quick reference to the original data frame that was returned from the
        app's render function. If it is mutated in place, it **will** modify the original
        data.

        Returns
        -------
        :
            Narwhals data frame object wrapping the original data.
        """
        return as_data_frame(self.data())

    def _nw_data_to_original_type(
        self,
        nw_data: DataFrame[IntoDataFrameT],
    ) -> IntoDataFrameT:
        """
        Convert the narwhals data frame back to the original data type.

        If the original data type was a `pandas` or `polars` data frame, this method
        will convert the narwhals data frame back to the original data type. However, if
        the original data type was a `narwhals` data frame, this method will maintain and return the narwhals data frame.

        Parameters
        ----------
        nw_data
            The narwhals data frame to convert.

        Returns
        -------
        :
            The original data frame type. Ex: pandas, polars, or narwhals
        """

        if isinstance(self.data(), DataFrame):
            # At this point, `IntoDataFrameT` represents `DataFrame[IntoDataFrameT]`
            return cast(IntoDataFrameT, nw_data)

        return data_frame_to_native(nw_data)

    _patch_fn: PatchFn
    """
    User-defined function to update a single cell in the data frame.

    The function takes a single keyword argument `patch` and (by default) returns the
    patch's value.
    """
    _patches_fn: PatchesFn
    """
    User-defined function to update all cells in a batch actions.

    It gives the user opportunity to return less, the same, or even more patches than
    originally requested by the browser.

    Defaults to calling `._patch_fn()` on each input patch and returning the input
    patches with updated values.
    """

    _cell_patch_map: reactive.Value[dict[tuple[int, int], CellPatch]]
    """
    Reactive dictionary of patches to be applied to the data frame.

    This map is used for faster deduplications of patches at each location given the row
    and column indices.

    The key is defined as `(row_index, column_index)`.
    """

    @reactive_calc_method
    def cell_patches(self) -> list[CellPatch]:
        """
        Reactive calculation of the data frame's edits.

        This reactive calculation that returns a list of user edits (or updated cell
        values) that will be applied to `.data()` to create the `.data_patched()` data
        frame.

        Returns
        -------
        :
            A list of cell patches to apply to the data frame.
        """
        return list(self._cell_patch_map().values())

    @reactive_calc_method
    def _nw_data_patched(self) -> DataFrame[IntoDataFrameT]:
        """
        Reactive calculation of the data frame's patched data.

        Returns
        -------
        :
            The data frame with all the user's edit patches applied to it.
        """
        return apply_frame_patches(self._nw_data(), self.cell_patches())

    @reactive_calc_method
    def data_patched(self) -> IntoDataFrameT:
        """
        Reactive calculation of the data frame's patched data.

        This method returns the `.data()` data frame with all `.cell_patches()` patches
        applied.

        The returned value is a shallow copy of the original data frame. It is possible
        that alterations to the `.data_patched()` data frame could alter other
        associated data frame values. Please be cautious when using this value directly.

        Returns
        -------
        :
            The patched data frame.
        """
        return self._nw_data_to_original_type(self._nw_data_patched())

    # Apply filtering and sorting
    # https://github.com/posit-dev/py-shiny/issues/1240
    def _subset_data_view(self, selected: bool) -> IntoDataFrameT:
        """
        Helper method to subset data according to what is viewed in the browser;

        Applies filtering and sorting to the patched data. If `selected=True`, only
        the user selected rows are returned.


        Note Future rect selection changes
        ----------------------------------
        In the future, the selected rows may need to be **after** filtering
        and sorting are applied. This would allow for rectangular selections to be
        applied to the filtered and sorted data given min/max row info.

        Where as, currently, the selected rows are applied to the original data
        before filtering and sorting are applied. Serializing the rect selection
        would require tuple info of all cells selected.

        Parameters
        ----------
        selected
            If `True`, subset the viewed data to the selected area. Defaults to `False` (all rows).

        Returns
        -------
        :
            The (possibly selected-only) subsetted data frame with any sorting or
            filtering applied. The data frame is of the same type as the render method's
            data.
        """

        if selected:
            rows = self.cell_selection()["rows"]
        else:
            rows = self.data_view_rows()

        nw_data = subset_frame(self._nw_data_patched(), rows=rows)

        patched_subsetted_into_data = self._nw_data_to_original_type(nw_data)

        return patched_subsetted_into_data

    @reactive_calc_method
    def _data_view_all(self) -> IntoDataFrameT:
        """
        Reactive calculation of the full (sorted and filtered) data.

        Returns
        -------
        :
            The full data frame as seen in the browser.
        """
        return self._subset_data_view(selected=False)

    @reactive_calc_method
    def _data_view_selected(self) -> IntoDataFrameT:
        """
        Reactive calcuation of the selected rows of the (sorted and filtered) data.

        Returns
        -------
        :
            The selected rows of the data frame as seen in the browser.
        """
        return self._subset_data_view(selected=True)

    @add_example(ex_dir="../api-examples/data_frame_data_view")
    def data_view(self, *, selected: bool = False) -> IntoDataFrameT:
        """
        Reactive function that retrieves the data how it is viewed within the browser.

        This function will return the `.data_patched()` data frame with the user's
        sorting and filtering applied. It represents the data frame as viewed by the
        user within the browser.

        The returned value is a shallow copy of the original data frame. It is possible
        that alterations to the `.data_view()` data frame could alter other associated
        date frame values. Please be cautious when using this value directly.

        Parameters
        ----------
        selected
            If `True`, subset the viewed data to the selected area. Defaults to `False`
            (all rows).

        Returns
        -------
        :
            A view of the (possibly selected) data frame as seen in the browser.

        See Also
        --------
        * [`pandas.DataFrame.copy` API documentation](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.copy.html)
        * [`polars.DataFrame.clone` API documentation](https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.clone.html)
        * [`narwhals.DataFrame.clone` API documentation](https://narwhals-dev.github.io/narwhals/api-reference/dataframe/#narwhals.dataframe.DataFrame.clone)
        """
        # Return reactive calculations so that they can be cached for other calculations
        if selected:
            return self._data_view_selected()
        else:
            return self._data_view_all()

    # @reactive_calc_method
    def selection_modes(self) -> SelectionModes:
        """
        Reactive calculation of the data frame's possible selection modes.

        Returns
        -------
        :
            The possible selection modes for the data frame.
        """
        return self._req_value().selection_modes

    @reactive_calc_method
    def cell_selection(self) -> CellSelection:
        """
        Reactive calculation of selected cell information.

        This method is a wrapper around `input.<id>_cell_selection()`, where `<id>` is
        the `id` of the data frame output. This method returns the selected rows and
        will cause reactive updates as the selected rows change.

        The value has been enhanced from it's vanilla form to include the missing `cols` key
        (or `rows` key) as a tuple of integers representing all column (or row) numbers.
        This allows for consistent usage within code when subsetting your data. These
        missing keys are not sent over the wire as they are independent of the selection.

        Returns
        -------
        :
            :class:`~shiny.render.CellSelection` representing the indices of the selected
            cells. If no cells are currently selected, `None` is returned.
        """
        browser_cell_selection = cast(
            BrowserCellSelection,
            self._get_session().input[f"{self.output_id}_cell_selection"](),
        )

        cell_selection = as_cell_selection(
            browser_cell_selection,
            selection_modes=self.selection_modes(),
            nw_data=self._nw_data(),
            data_view_rows=self.data_view_rows(),
            data_view_cols=tuple(range(self._nw_data().shape[1])),
        )

        return cell_selection

    @reactive_calc_method
    def data_view_rows(self) -> tuple[int, ...]:
        """
        Reactive calculation of the data frame's user view row numbers.

        This value is a wrapper around `input.<id>_data_view_rows()`, where `<id>` is the
        `id` of the data frame output.

        Returns
        -------
        :
            The row numbers of the data frame that are currently being viewed in the browser
            after sorting and filtering has been applied.
        """
        input_data_view_rows = self._get_session().input[
            f"{self.output_id}_data_view_rows"
        ]()
        return tuple(input_data_view_rows)

    # @reactive_calc_method
    def sort(self) -> tuple[ColumnSort, ...]:
        """
        Reactive calculation of the data frame's column sorting information.

        Returns
        -------
        :
            An array of `col`umn number and _is `desc`ending_ information.
        """
        column_sort = self._get_session().input[f"{self.output_id}_column_sort"]()
        return tuple(column_sort)

    # @reactive_calc_method
    def filter(self) -> tuple[ColumnFilter, ...]:
        """
        Reactive calculation of the data frame's column filters.

        Returns
        -------
        :
            An array of `col`umn number and `value` information. If the column type is a number, a tuple of `(min, max)` is used for `value`. If no min (or max) value is set, `None` is used in its place. If the column type is a string, the string value is used for `value`.
        """
        column_filter = self._get_session().input[f"{self.output_id}_column_filter"]()
        return tuple(column_filter)

    def _reset_reactives(self) -> None:
        self._value.set(None)
        self._cell_patch_map.set({})
        self._updated_data.unset()

    def _init_reactives(self) -> None:

        # Init
        self._value = reactive.Value(None)
        self._cell_patch_map = reactive.Value({})
        self._updated_data = reactive.Value()  # Create with no value

        # Update the styles any time the cell patch map or new data updates
        def should_update_styles():
            return (
                self._cell_patch_map(),
                # If the udpated data is unset, use a `None` value which is not allowed.
                self._updated_data() if self._updated_data.is_set() else None,
            )

        @reactive.effect
        @reactive.event(
            should_update_styles,
            # Do not run the first time through!
            # The styles are being sent with the initial blob.
            ignore_init=True,
        )
        async def _():
            # Be sure this is called within `isolate()` to isolate any reactivity
            # It currently is, as `@reactive.event()` is being used
            await self._attempt_update_cell_style()

    def _get_session(self) -> Session:
        if self._session is None:
            raise RuntimeError(
                "The data frame being used was not initialized within a reactive context / session. Please call `@render.data_frame` where `shiny.session.get_current_context()` returns a non-`None` value."
            )
        return self._session

    @add_example(ex_dir="../api-examples/data_frame_data_view")
    def set_patch_fn(self, fn: PatchFn | PatchFnSync) -> None:
        """
        Decorator to set the function that updates a single cell in the data frame.

        The default patch function returns the value as is.

        Parameters
        ----------
        fn
            A function that accepts a kwarg `patch` and returns the processed
            `patch.value` for the cell.
        """
        self._patch_fn = wrap_async(  # pyright: ignore[reportGeneralTypeIssues,reportAttributeAccessIssue]
            fn
        )
        # Do not return self here as it is typically used as a decorator (which would return `self`)
        # By returning `self` in express mode, it is attempted to be registered twice. That is bad.
        # So for now, we will not return `self` here.
        # from .._typing_extensions import Self
        # return self

    @add_example(ex_dir="../api-examples/data_frame_set_patches")
    def set_patches_fn(self, fn: PatchesFn | PatchesFnSync) -> None:
        """
        Decorator to set the function that updates a batch of cells in the data frame.

        The default patches function calls the async `._patch_fn()` on each input patch
        and returns the updated patch values.

        There are no checks made on the quantity of patches returned. The user can
        return more, less, or the same number of patches as the input patches. This
        allows for the app author to own more control over which columns are updated and
        how they are updated.

        Parameters
        ----------
        fn
            A function that accepts a kwarg `patches` and returns a list of (possibly
            updated) patches to apply to the data frame.
        """
        self._patches_fn = wrap_async(  # pyright: ignore[reportGeneralTypeIssues,reportAttributeAccessIssue]
            fn
        )
        # Do not return self here as it is typically used as a decorator (which would return `self`)
        # By returning `self` in express mode, it is attempted to be registered twice. That is bad.
        # So for now, we will not return `self` here.
        # from .._typing_extensions import Self
        # return self

    def _init_patch_fns(self) -> None:
        """
        Initialize `._patch_fn()` and `._patches_fn()`.
        """

        async def patch_fn(
            *,
            patch: CellPatch,
        ) -> CellValue:
            return patch["value"]

        async def patches_fn(
            *,
            patches: tuple[CellPatch, ...],
        ) -> ListOrTuple[CellPatch]:
            ret_patches: list[CellPatch] = []
            for patch in patches:

                new_patch = patch.copy()
                new_patch["value"] = await self._patch_fn(patch=patch)
                ret_patches.append(new_patch)

            return ret_patches

        self.set_patch_fn(patch_fn)
        self.set_patches_fn(patches_fn)

    def _set_patches_handler_impl(
        self,
        handler: Callable[..., Awaitable[Jsonifiable]] | None,
    ) -> str:
        """
        Set the client patches request handler for the data frame.

        This method should be removed when the rendered result is `None`.
        (... b/c there is no data frame to send requests!)
        """
        session = self._get_session()
        key = session.set_message_handler(
            f"data_frame_patches_{self.output_id}",
            handler,
        )
        return key

    def _reset_patches_handler(self) -> str:
        """
        Resets the client patches request handler for the data frame.
        """
        return self._set_patches_handler_impl(None)

    def _set_patches_handler(self) -> str:
        """
        Set the client patches handler for the data frame.

        This method **must be** called as late as possible as it depends on the ID of the output.
        """
        return self._set_patches_handler_impl(self._patches_handler)

    # Do not change this method name unless you update corresponding code in `/js/dataframe/`!!
    async def _patches_handler(self, patches: tuple[CellPatch, ...]) -> Jsonifiable:
        """
        Accepts edit patches requests from the client and returns the processed patches.

        Parameters
        ----------
        patches
            A list of patches to apply to the data frame.

        Returns
        -------
        :
            A list of processed patches to apply to the data frame. The number of
            processed patches can be different from the number of input patches.
        """
        assert_patches_shape(patches)

        with session_context(self._get_session()):
            # Call user's cell update method to retrieve formatted values
            val = await self._patches_fn(patches=patches)
            patches = tuple(val)

        # Check to make sure `updated_infos` is a list of dicts with the correct keys
        bad_patches_format = not isinstance(patches, list)
        if not bad_patches_format:
            for patch in patches:
                if not (
                    # Verify structure
                    isinstance(patch, dict)
                    # Verify types
                    and isinstance(patch["row_index"], int)
                    and isinstance(patch["column_index"], int)
                    # # Do not check the value type here. It should be validated by
                    # # `._set_cell_patch_map_patches()` later with more type hint context
                    # and isinstance(updated_patch["value"], CellValue)
                ):
                    raise ValueError(
                        f"The return value of {self.output_id}'s `_patches_fn()` "
                        f"(typically set by `@{self.output_id}.set_patches_fn`) "
                        f"must be a list where each item has a row_index (`int`), column_index (`int`), and value (`TagChild`)."
                    )

        # Add (or overwrite) new cell patches by setting each patch into the cell patch map
        jsonifiable_processed_patches = self._set_cell_patch_map_patches(patches)

        # Return the processed patches to the client
        return jsonifiable_processed_patches

    def _set_cell_patch_map_patches(
        self,
        patches: ListOrTuple[CellPatch],
    ) -> list[Jsonifiable]:
        """
        Set the patches within the cell patch map.

        Parameters
        ----------
        patches
            Set of patches to apply to store in the cell patch map.

        Returns
        -------
        :
            A list of processed (by the session) patches to apply to the data frame.
        """
        # Use copy to set the new value
        cell_patch_map = self._cell_patch_map().copy()

        for patch in patches:
            row_index = patch["row_index"]
            column_index = patch["column_index"]

            assert isinstance(
                row_index, int
            ), f"Expected `row_index` to be an `int`, received {type(row_index)}"
            assert isinstance(
                column_index, int
            ), f"Expected `column_index` to be an `int`, received {type(column_index)}"

            # TODO-render.data_frame; Possibly check for cell type and compare against type hints
            # TODO-render.data_frame; The `value` should be coerced by pandas to the correct type
            # TODO-render.data_frame; See https://pandas.pydata.org/pandas-docs/stable/user_guide/basics.html#object-conversion

            cell_patch_map[(row_index, column_index)] = patch

        # Once all patches are set, update the cell patch map with new version
        self._cell_patch_map.set(cell_patch_map)

        # Upgrade any HTML-like content to `CellHtml` json objects
        # for sending to the client
        session = self._get_session()
        processed_patches: list[CellPatchProcessed] = [
            {
                "row_index": patch["row_index"],
                "column_index": patch["column_index"],
                # Only upgrade the value if it is necessary
                "value": maybe_as_cell_html(
                    patch["value"],
                    session=session,
                ),
            }
            for patch in patches
        ]

        # Prep the processed patches as dictionaries for sending to the client
        jsonifiable_patches: list[Jsonifiable] = [
            cell_patch_processed_to_jsonifiable(ret_processed_patch)
            for ret_processed_patch in processed_patches
        ]

        return jsonifiable_patches

    async def _attempt_update_cell_style(self) -> None:

        rendered_value = self._value()
        if not isinstance(rendered_value, (DataGrid, DataTable)):
            return

        styles_fn = rendered_value.styles
        if not callable(styles_fn):
            return

        patched_into_data = self._nw_data_to_original_type(self._nw_data_patched())
        new_styles = as_browser_style_infos(styles_fn, into_data=patched_into_data)

        await self._send_message_to_browser(
            "updateStyles",
            {"styles": new_styles},
        )

    async def update_cell_value(
        self,
        value: CellValue,
        *,
        row: int,
        col: int | str,
    ) -> None:
        """
        Update the value of a cell in the data frame.

        Calling this method will set a new entry in `.cell_patches()`. It will not reset
        the user's sorting or filtering of their rendered data frame.

        Parameters
        ----------
        value
            The new value to set the cell to.
        row
            The row index of the cell to update.
        column
            The column index of the cell to update.
        """
        # TODO-barret; Test these assertions
        # Convert column name to index if necessary
        if isinstance(col, str):
            with reactive.isolate():
                column_names = self._nw_data().columns
                if col not in column_names:
                    raise ValueError(f"Column '{col}' not found in data frame.")
                column_index = self._nw_data().columns.index(col)
        else:
            column_index = col
            if (not isinstance(col, int)) or isinstance(col, bool):
                raise TypeError(
                    f"Expected `col` to be an `int` or `str, received {type(column_index)}"
                )
            if column_index < 0:
                raise ValueError(
                    f"Expected `col` to be greater than or equal to 0, received {column_index}"
                )
            column_length = self._nw_data().shape[1]
            if column_index >= column_length:
                raise ValueError(
                    f"Expected `col` to be less than {column_length}, received {column_index}"
                )

        if (not isinstance(row, int)) or isinstance(row, bool):
            raise TypeError(f"Expected `row` to be an `int`, received {type(row)}")
        if row < 0:
            raise ValueError(
                f"Expected `row` to be greater than or equal to 0, received {row}"
            )

        patch: CellPatch = {
            "value": value,
            "row_index": row,
            "column_index": column_index,
        }
        processed_patches = self._set_cell_patch_map_patches([patch])

        # TODO-barret-render.data_frame; Send message to client to update cell value
        await self._send_message_to_browser(
            "addPatches",
            {"patches": processed_patches},
        )

        return

    async def update_data(
        self,
        data: IntoDataFrameT,
    ) -> None:
        """
        Update the data frame with new data.

        Calling this method will update the `.data()` data frame with new data and will
        remove all `.cell_patches()`. It will not reset the user's sorting or filtering
        of their rendered data frame. Any incompatible sorting or filtering settings
        will be silently dropped.

        Parameters
        ----------
        data
            The new data to render.
        """
        assert_data_is_not_none(data)

        # Serialize the data within the session context,
        # similar to `.to_payload()` on the `._value()`
        with session_context(self._get_session()):
            info = serialize_frame(data)

        # Reset patches & set new data
        # Perform only after serializing the frame
        # (which performs sanity checks. E.g. `data is not None`
        self._cell_patch_map.set({})
        self._updated_data.set(data)

        await self._send_message_to_browser(
            "updateData",
            {
                "data": info["data"],
                "columns": info["columns"],
                "typeHints": info["typeHints"],
            },
        )
        return

    def auto_output_ui(self) -> Tag:
        return ui.output_data_frame(id=self.output_id)

    def __init__(
        self,
        fn: ValueFn[
            None | IntoDataFrameT | DataGrid[IntoDataFrameT] | DataTable[IntoDataFrameT]
        ],
    ):
        super().__init__(fn)

        # Set reactives from calculated properties
        self._init_reactives()
        # Set update functions
        self._init_patch_fns()

    def _set_output_metadata(self, *, output_id: str) -> None:
        super()._set_output_metadata(output_id=output_id)

        # Verify that the session used (during `__init__`) when creating the renderer is
        # the same session used when executing the renderer. This is to prevent a user
        # from creating a renderer in one module and registering it on an output with a
        # different session.

        active_session = require_active_session(None)
        if self._get_session() != active_session:
            raise RuntimeError(
                "The session used when creating the renderer "
                "is not the same session used when executing the renderer. "
                "Please file an issue on "
                "GitHub <https://github.com/posit-dev/py-shiny/issues/new> "
                "with an example of how you are reproducing this error. "
                "We would be curious to know your use case!"
            )

    async def render(self) -> JsonifiableDict | None:
        # Reset value
        self._reset_reactives()
        self._reset_patches_handler()

        value = await self.fn()
        if value is None:
            return None

        if not isinstance(value, (DataGrid, DataTable)):
            try:
                value = DataGrid(value)
            except TypeError as e:
                raise TypeError(
                    "@render.data_frame doesn't know how to render objects of type ",
                    type(value),
                ) from e

        # Set patches url handler for client
        patch_key = self._set_patches_handler()

        self._value.set(value)  # pyright: ignore[reportArgumentType]

        # Use session context so `to_payload()` gets the correct session
        with session_context(self._get_session()):
            payload = value.to_payload()

            ret: FrameRender = {
                "payload": payload,
                "patchInfo": {
                    "key": patch_key,
                },
                "selectionModes": self.selection_modes().as_dict(),
            }
            return frame_render_to_jsonifiable(ret)

    async def _send_message_to_browser(self, handler: str, obj: dict[str, Any]):

        session = self._get_session()
        await session.send_custom_message(
            "shinyDataFrameMessage",
            {
                # Custom message handlers are never namespaced for modules.
                # Must provide a unique name to avoid conflicts between modules.
                "id": session.ns(self.output_id),
                "handler": handler,
                "obj": obj,
            },
        )

    async def update_cell_selection(
        # self, selection: SelectionLocation | CellSelection
        self,
        selection: CellSelection | Literal["all"] | None | BrowserCellSelection,
    ) -> None:
        """
        Update the cell selection in the data frame.

        Currently only single (`"type": "row"`) or multiple (`"type": "rows"`) row
        selection is supported.

        If the current data frame selection mode is `"none"` and a non-none selection is
        provided, a warning will be raised and no rows will be selected. If cells are
        supposes to be selected, the selection mode returned from the render function
        must (currently) be set to `"row"` or `"rows"`.

        Parameters
        ----------
        selection
            The cell selection to apply to the data frame. This can be a `CellSelection`
            object, `"all"` to select all cells (if possible), or `None` to clear the
            selection.
        """
        with reactive.isolate():
            selection_modes = self.selection_modes()
            nw_data = self._nw_data()
            data_view_rows = self.data_view_rows()
            data_view_cols = tuple(range(nw_data.shape[1]))

        if selection_modes._is_none():
            warnings.warn(
                'Cell selection cannot be updated when `.selection_mode=` contains "none". '
                'Please set `selection_mode=` to contain a non-`"none"` value '
                "(e.g. 'row' or 'rows') in the return value of "
                "`@render.data_frame` to enable cell selection.",
                stacklevel=2,
            )

            selection = None

        cell_selection = as_cell_selection(
            selection,
            selection_modes=selection_modes,
            nw_data=nw_data,
            data_view_rows=data_view_rows,
            data_view_cols=data_view_cols,
        )

        if cell_selection["type"] == "none":
            pass
        elif cell_selection["type"] == "rect":
            raise RuntimeError("Rectangle region selection is not yet supported")
        elif cell_selection["type"] == "col":
            raise RuntimeError("Column selection is not yet supported")
        elif cell_selection["type"] == "row":
            row_value = cell_selection["rows"]
            if selection_modes.row == "single" and len(row_value) > 1:
                warnings.warn(
                    "Attempted to set cell selection to more than 1 row when `.selection_modes()` contains 'row'. "
                    "Only the first row supplied will be selected.",
                    stacklevel=2,
                )

                cell_selection["rows"] = (row_value[0],)
        else:
            raise ValueError(f"Unhandled selection type: {cell_selection['type']}")
        await self._send_message_to_browser(
            "updateCellSelection",
            {"cellSelection": cell_selection},
        )

    @add_example(ex_dir="../api-examples/data_frame_update_sort")
    async def update_sort(
        self,
        sort: ListOrTuple[ColumnSort | int] | int | ColumnSort | None,
    ) -> None:
        """
        Update the column sorting in the data frame.

        The sort will be applied in reverse order so that the first value has the highest
        precedence. This mean _ties_ will go to the second sort column (and so on).

        Parameters
        ----------
        sort
            A list of column sorting information. If `None`, sorting will be removed. `int` values will be upgraded to `{"col": int, "desc": <DESC>}` where `<DESC>` is `True` if the column is number like and `False` otherwise.
        """
        if sort is None:
            sort = ()
        elif isinstance(sort, int):
            sort = (sort,)
        elif isinstance(sort, (list, tuple)):
            ...
        else:
            raise TypeError(
                f"Expected `sort` to be a `list`, `tuple`, `int`, or `None`. Received `{type(sort)}`"
            )

        vals: list[ColumnSort] = []
        if len(sort) > 0:
            with reactive.isolate():
                nw_data = self._nw_data()
            ncol = nw_data.shape[1]

            for val in sort:
                val_dict: ColumnSort
                if isinstance(val, int):
                    desc = nw_data[:, val].dtype.is_numeric()
                    val_dict = {"col": val, "desc": desc}
                val_dict: ColumnSort = (
                    val if isinstance(val, dict) else {"col": val, "desc": True}
                )
                assert isinstance(val_dict, dict)
                assert isinstance(val_dict["col"], int)
                assert 0 <= val_dict["col"] < ncol
                assert isinstance(val_dict["desc"], bool)
                vals.append(val_dict)

        await self._send_message_to_browser(
            "updateColumnSort",
            {"sort": vals},
        )

    @add_example(ex_dir="../api-examples/data_frame_update_filter")
    async def update_filter(
        self,
        filter: ListOrTuple[ColumnFilter] | None,
    ) -> None:
        """
        Update the column filtering in the data frame.

        Parameters
        ----------
        filter
            A list of column filtering information. If `None`, filtering will be removed.
        """

        assert filter is None or isinstance(filter, (list, tuple))

        if filter is None:
            filter = []
        else:
            with reactive.isolate():
                ncol = self._nw_data().shape[1]

            for column_filter, i in zip(filter, range(len(filter))):
                assert isinstance(column_filter, dict)
                assert isinstance(column_filter["col"], int)
                assert 0 <= column_filter["col"] < ncol
                if isinstance(column_filter["value"], str):
                    ...
                elif isinstance(column_filter["value"], (list, tuple)):
                    assert len(column_filter["value"]) == 2
                    if (
                        column_filter["value"][0] is None
                        and column_filter["value"][1] is None
                    ):
                        raise TypeError(
                            "Expected `filter[{i}]['value']` to be a `str` or a `list`/`tuple` of type `int` or `None`. Received `None` for both values."
                        )
                    assert isinstance(
                        column_filter["value"][0], (int, float, type(None))
                    )
                    assert isinstance(
                        column_filter["value"][1], (int, float, type(None))
                    )
                else:
                    raise TypeError(
                        f"Expected `filter[{i}]['value']` to be a `str` or a `list`/`tuple` of type `int` or `None`. Received `{type(column_filter['value'])}`"
                    )

        await self._send_message_to_browser(
            "updateColumnFilter",
            {"filter": filter},
        )

    def input_cell_selection(self) -> CellSelection | None:
        """
        [Deprecated] Reactive calculation of selected cell information.

        Please use `~shiny.render.data_frame`'s `.cell_selection()` method instead.
        """

        from .._deprecated import warn_deprecated

        warn_deprecated(
            "`@render.data_frame`'s `.input_cell_selection()` method is deprecated. Please use `.cell_selection()` instead."
        )

        cell_selection = self.cell_selection()
        if cell_selection["type"] == "none":
            return None

        return self.cell_selection()


# TODO-barret; Make request to GT: Add class for gt location

# TODO-barret; Are GT formatters eager or lazy?
# GT styles are eager! THis causes issues with dynamic style locations
# Need to track movement of columns

# SKIP!: ONLY SUPPORT GT and editable==False; Sprint for next 24 hrs!

# THIS v1 release;
# * styles= Callable[[data], Styles] | Styles
# * styles function will return json format which is
#   Styles = list[StyleInfo]
#   StyleInfo = {loc: "body", rows: int | List[int], columns: int | str | List[int | str], style: str? | dict[str, Jsonifiable]? | None, class_: str | None, }
# * Call styles fn after each edit (and on init); Send full styles for each cell
# Support `rows: List[bool]`?

# mydf.iloc[row_nums, col_nums]
# mydf["mpg"] > 20

# myints = [i for i, val in enumerate(range(mybools)) if val]
# mybools = [val in myints for i, val in enumerate(range(nrow))]



from shiny.express import expressify, ui


# @expressify converts a function to work with Express syntax
@expressify
def expressified1(s: str):
    f"Expressified function 1: {s}"  # noqa: B021
    ui.br()


expressified1("Hello")

expressified1("world")


ui.br()


# @expressify() also works with parens
@expressify()
def expressified2(s: str):
    f"Expressified function 2: {s}"  # noqa: B021
    ui.br()


expressified2("Hello")

expressified2("world")



from datetime import date

from shiny import App, Inputs, Outputs, Session, reactive, ui

app_ui = ui.page_fluid(
    ui.panel_title("Changing the values of inputs from the server"),
    ui.row(
        ui.column(
            4,
            ui.panel_well(
                ui.tags.h4("These inputs control the other inputs on the page"),
                ui.input_text(
                    "control_label", "This controls some of the labels:", "LABEL TEXT"
                ),
                ui.input_slider(
                    "control_num", "This controls values:", min=1, max=20, value=15
                ),
            ),
        ),
        ui.column(
            4,
            ui.panel_well(
                ui.tags.h4("These inputs are controlled by the other inputs"),
                ui.input_text("inText", "Text input:", value="start text"),
                ui.input_numeric(
                    "inNumber", "Number input:", min=1, max=20, value=5, step=0.5
                ),
                ui.input_numeric(
                    "inNumber2", "Number input 2:", min=1, max=20, value=5, step=0.5
                ),
                ui.input_slider("inSlider", "Slider input:", min=1, max=20, value=15),
                ui.input_slider(
                    "inSlider2", "Slider input 2:", min=1, max=20, value=(5, 15)
                ),
                ui.input_slider(
                    "inSlider3", "Slider input 3:", min=1, max=20, value=(5, 15)
                ),
                ui.input_date("inDate", "Date input:"),
                ui.input_date_range("inDateRange", "Date range input:"),
            ),
        ),
        ui.column(
            4,
            ui.panel_well(
                ui.input_checkbox("inCheckbox", "Checkbox input", value=False),
                ui.input_checkbox_group(
                    "inCheckboxGroup",
                    "Checkbox group input:",
                    {
                        "option1": "label 1",
                        "option2": "label 2",
                    },
                ),
                ui.input_radio_buttons(
                    "inRadio",
                    "Radio buttons:",
                    {
                        "option1": "label 1",
                        "option2": "label 2",
                    },
                ),
                ui.input_select(
                    "inSelect",
                    "Select input:",
                    {
                        "option1": "label 1",
                        "option2": "label 2",
                    },
                ),
                ui.input_select(
                    "inSelect2",
                    "Select input 2:",
                    {
                        "option1": "label 1",
                        "option2": "label 2",
                    },
                    multiple=True,
                ),
            ),
            ui.navset_tab(
                ui.nav_panel("panel1", ui.h2("This is the first panel.")),
                ui.nav_panel("panel2", ui.h2("This is the second panel.")),
                id="inTabset",
            ),
        ),
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.effect
    def _():
        # We'll use these multiple times, so use short var names for
        # convenience.
        c_label = input.control_label()
        c_num = input.control_num()

        # Text =====================================================
        # Change both the label and the text
        ui.update_text(
            "inText",
            label="New " + c_label,
            value="New text " + str(c_num),
        )

        # Number ===================================================
        # Change the value
        ui.update_numeric("inNumber", value=c_num)

        # Change the label, value, min, and max
        ui.update_numeric(
            "inNumber2",
            label="Number " + c_label,
            value=c_num,
            min=c_num - 10,
            max=c_num + 10,
            step=5,
        )

        # Slider input =============================================
        # Only label and value can be set for slider
        ui.update_slider("inSlider", label="Slider " + c_label, value=c_num)

        # Slider range input =======================================
        # For sliders that pick out a range, pass in a vector of 2
        # values.
        ui.update_slider("inSlider2", value=(c_num - 1, c_num + 1))

        # Only change the upper handle
        ui.update_slider("inSlider3", value=(input.inSlider3()[0], c_num + 2))

        # Date input ===============================================
        # Only label and value can be set for date input
        ui.update_date("inDate", label="Date " + c_label, value=date(2013, 4, c_num))

        # Date range input =========================================
        # Only label and value can be set for date range input
        ui.update_date_range(
            "inDateRange",
            label="Date range " + c_label,
            start=date(2013, 1, c_num),
            end=date(2013, 12, c_num),
            min=date(2001, 1, c_num),
            max=date(2030, 1, c_num),
        )

        # # Checkbox ===============================================
        ui.update_checkbox("inCheckbox", value=c_num % 2)

        # Checkbox group ===========================================
        # Create a list of new options, where the name of the items
        # is something like 'option label x A', and the values are
        # 'option-x-A'.
        opt_labels = [f"option label {c_num} {type}" for type in ["A", "B"]]
        opt_vals = [f"option-{c_num}-{type}" for type in ["A", "B"]]
        opts_dict = dict(zip(opt_vals, opt_labels))

        # Set the label, choices, and selected item
        ui.update_checkbox_group(
            "inCheckboxGroup",
            label="Checkbox group " + c_label,
            choices=opts_dict,
            selected=f"option-{c_num}-A",
        )

        # Radio group ==============================================
        ui.update_radio_buttons(
            "inRadio",
            label="Radio " + c_label,
            choices=opts_dict,
            selected=f"option-{c_num}-A",
        )
        # Select input =============================================
        # Create a list of new options, where the name of the items
        # is something like 'option label x A', and the values are
        # 'option-x-A'.
        ui.update_select(
            "inSelect",
            label="Select " + c_label,
            choices=opts_dict,
            selected=f"option-{c_num}-A",
        )

        # Can also set the label and select an item (or more than
        # one if it's a multi-select)
        ui.update_select(
            "inSelect2",
            label="Select label " + c_label,
            choices=opts_dict,
            selected=f"option-{c_num}-B",
        )

        # Tabset input =============================================
        # Change the selected tab.
        # The tabsetPanel must have been created with an 'id' argument
        ui.update_navs("inTabset", selected="panel2" if c_num % 2 else "panel1")


app = App(app_ui, server, debug=True)



# --------------------------------------------------------------------------------
# This example demonstrates how to clear the chat when the model changes.
# To run it, you'll need an OpenAI API key.
# To get one, follow the instructions at https://platform.openai.com/docs/quickstart
# --------------------------------------------------------------------------------
import os

from langchain_openai import ChatOpenAI

from shiny import reactive
from shiny.express import input, ui

# Provide your API key here (or set the environment variable)
llm = ChatOpenAI(
    api_key=os.environ.get("OPENAI_API_KEY"),  # type: ignore
)

# Set some Shiny page options
ui.page_opts(
    title="Hello OpenAI Chat",
    fillable=True,
    fillable_mobile=True,
)

# Create a sidebar to select the model
with ui.sidebar():
    ui.input_select("model", "Model", ["gpt-4o", "gpt-3.5-turbo"])

# Create and display an empty chat UI
chat = ui.Chat(id="chat")
chat.ui()


# Define a callback to run when the user submits a message
@chat.on_user_submit
async def _():
    # Get messages currently in the chat
    messages = chat.messages(format="langchain")
    # Create a response message stream
    response = llm.astream(messages)
    # Append the response stream into the chat
    await chat.append_message_stream(response)


# Clear the chat when the model changes
@reactive.effect
@reactive.event(input.model)
async def _():
    await chat.clear_messages()



from conftest import create_doc_example_core_fixture
from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc

app = create_doc_example_core_fixture("update_action_button")


def test_input_action_kitchen(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    button2 = controller.InputActionButton(page, "goButton2")
    expect(button2.loc).to_have_text("🤩 Go 2")
    button2.expect_label("🤩 Go 2")
    button2.expect_width(None)

    link = controller.InputActionLink(page, "goLink")
    link.expect_label("Go Link")

    button1 = controller.InputActionButton(page, "goButton")
    button1.expect_label("Go")
    button3 = controller.InputActionButton(page, "goButton3")
    button3.expect_label("Go 3")

    controller.InputActionButton(page, "update").click()

    button1.expect_label("📅 New label")
    button2.expect_label("🤩 Go 2")
    button3.expect_label("New label 3")
    link.expect_label("🔗 New link label")



from shiny import App, Inputs, Outputs, Session, reactive, ui

app_ui = ui.page_fluid(
    ui.layout_sidebar(
        ui.sidebar(
            ui.input_radio_buttons(
                "controller", "Controller", ["1", "2", "3"], selected="1"
            )
        ),
        ui.navset_hidden(
            ui.nav_panel(None, "Panel 1 content", value="panel1"),
            ui.nav_panel(None, "Panel 2 content", value="panel2"),
            ui.nav_panel(None, "Panel 3 content", value="panel3"),
            id="hidden_tabs",
        ),
    )
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.effect
    @reactive.event(input.controller)
    def _():
        ui.update_navs("hidden_tabs", selected="panel" + str(input.controller()))


app = App(app_ui, server)



from __future__ import annotations

from htmltools import HTMLDependency

from .._versions import bslib as bslib_version
from .._versions import htmltools as htmltools_version

"""
HTML dependencies for shinyverse dependencies from R packages such as bslib or htmltools.

For...
* External dependencies (e.g. jQuery, Bootstrap), see `shiny.ui._html_deps_external`
* Internal dependencies (e.g. dataframe, autoresize), see `shiny.ui._html_deps_py_shiny`
* shinyverse dependencies (e.g. bslib, htmltools), see `shiny.ui._html_deps_shinyverse`
"""


_rel_www_shared = "www/shared"
_htmltools_path = f"{_rel_www_shared}/htmltools"
_components_path = f"{_rel_www_shared}/bslib/components"


# -- htmltools ---------------------


def fill_dependency() -> HTMLDependency:
    return HTMLDependency(
        name="htmltools-fill",
        version=htmltools_version,
        source={
            "package": "shiny",
            "subdir": f"{_htmltools_path}/fill",
        },
        stylesheet={"href": "fill.css"},
    )


# -- bslib -------------------------


def components_dependencies(include_css: bool = True) -> HTMLDependency:
    return HTMLDependency(
        name="bslib-components",
        version=bslib_version,
        source={
            "package": "shiny",
            "subdir": f"{_components_path}",
        },
        script=[
            {"src": "components.min.js"},
            {"src": "web-components.min.js", "type": "module"},
        ],
        # This CSS is rendered against default Bootstrap and is only used in a
        # context where Shiny's Bootstrap is not used.
        stylesheet={"href": "components.css"} if include_css else None,
    )



from pathlib import Path

from shiny import App, ui

css_file = Path(__file__).parent / "css" / "styles.css"

app_ui = ui.page_fluid(
    "Almost before we knew it, we had left the ground!!!",
    ui.include_css(css_file),
    ui.div(
        # Style individual elements with an attribute dictionary.
        {"style": "font-weight: bold"},
        ui.p("Bold text"),
    ),
)

app = App(app_ui, None)



from __future__ import annotations

from pathlib import Path
from typing import Callable

from htmltools import Tag

from ... import ui
from ..._docstring import add_example
from ...types import MISSING, MISSING_TYPE
from ...ui._html_deps_external import ThemeProvider
from .._recall_context import RecallContextManager
from .._run import get_top_level_recall_context_manager

__all__ = ("page_opts",)


def page_auto_cm() -> RecallContextManager[Tag]:
    return RecallContextManager(ui.page_auto)


@add_example()
def page_opts(
    *,
    title: str | MISSING_TYPE = MISSING,
    window_title: str | MISSING_TYPE = MISSING,
    lang: str | MISSING_TYPE = MISSING,
    theme: str | Path | ui.Theme | ThemeProvider | MISSING_TYPE = MISSING,
    page_fn: Callable[..., Tag] | None | MISSING_TYPE = MISSING,
    fillable: bool | MISSING_TYPE = MISSING,
    full_width: bool | MISSING_TYPE = MISSING,
    **kwargs: object,
) -> None:
    """
    Set page-level options for the current app.

    The arguments to this function get passed to :func:`~shiny.ui.page_auto`, which
    determines which page function should be used based on the page options and the
    top-level items in the app.

    If there is a top-level :func:`~shiny.ui.nav_panel`, :func:`~shiny.ui.page_auto`
    will use :func:`~shiny.ui.page_navbar`. Otherwise, if there is a top-level sidebar,
    :func:`~shiny.ui.page_sidebar` is used.

    If there are neither top-level nav panels nor sidebars, this will use the `fillable`
    and `full_width` arguments to determine which page function to use:

    1. If `fillable` is `True`, :func:`~shiny.ui.page_fillable` is used.
    2. Otherwise, if `full_width` is `True`, :func:`~shiny.ui.page_fluid` is used.
    3. If neither are `True`, :func:`~shiny.ui.page_fixed` is used.

    Parameters
    ----------
    title
        A title shown on the page.
    window_title
        The browser window title. If no value is provided, this will use the value of
        ``title``.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This
        will be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The
        default, `None`, results in an empty string.
    theme
        A custom Shiny theme created using the :class:`~shiny.ui.Theme` class, or a path
        to a local or online CSS file that will replace the Bootstrap CSS bundled by
        default with a Shiny app. This file should be a complete `bootstrap.css` or
        `bootstrap.min.css` file.

        For advanced uses, you can also pass a :class:`~htmltools.Tagifiable` object.
        In this case, Shiny will suppress the default Bootstrap CSS.

        To modify the theme of an app without replacing the Bootstrap CSS entirely, use
        :func:`~shiny.ui.include_css` to add custom CSS.
    fillable
        If there is a top-level sidebar or nav, then the value is passed through to the
        :func:`~shiny.ui.page_sidebar` or :func:`~shiny.ui.page_navbar` function.
        Otherwise, if ``True``, use :func:`~shiny.ui.page_fillable`, where the content
        fills the window; if ``False`` (the default), the value of ``full_width`` will
        determine which page function is used.
    full_width
        This has an effect only if there are no sidebars or top-level navs, and
        ``fillable`` is ``False``. If this is ``False`` (the default), use use
        :func:`~shiny.ui.page_fixed`; if ``True``, use :func:`~shiny.ui.page_fillable`.
    page_fn
        The page function to use. If ``None`` (the default), will automatically choose
        one based on the arguments provided. If not ``None``, this will override all
        heuristics for choosing page functions.
    **kwargs
        Additional arguments to pass to the page function. See the description above for
        further details on how the page function is selected.
    """
    try:
        cm = get_top_level_recall_context_manager()
    except RuntimeError:
        # We can get here if a Shiny Core app, or if we're in the UI rendering phase of
        # a Quarto-Shiny dashboard.
        raise RuntimeError(
            "express.ui.page_opts() can only be used inside of a standalone Shiny Express app."
        )

    if not isinstance(title, MISSING_TYPE):
        cm.kwargs["title"] = title
    if not isinstance(window_title, MISSING_TYPE):
        cm.kwargs["window_title"] = window_title
    if not isinstance(lang, MISSING_TYPE):
        cm.kwargs["lang"] = lang
    if not isinstance(theme, MISSING_TYPE):
        cm.kwargs["theme"] = theme
    if not isinstance(page_fn, MISSING_TYPE):
        cm.kwargs["page_fn"] = page_fn
    if not isinstance(fillable, MISSING_TYPE):
        cm.kwargs["fillable"] = fillable
    if not isinstance(full_width, MISSING_TYPE):
        cm.kwargs["full_width"] = full_width
    if len(kwargs) > 0:
        cm.kwargs.update(kwargs)



import seaborn as sns
from faicons import icon_svg

# Import data from shared.py
from shared import app_dir, df

from shiny import reactive
from shiny.express import input, render, ui

ui.page_opts(title="Penguins dashboard", fillable=True)


with ui.sidebar(title="Filter controls"):
    ui.input_slider("mass", "Mass", 2000, 6000, 6000)
    ui.input_checkbox_group(
        "species",
        "Species",
        ["Adelie", "Gentoo", "Chinstrap"],
        selected=["Adelie", "Gentoo", "Chinstrap"],
    )


with ui.layout_column_wrap(fill=False):
    with ui.value_box(showcase=icon_svg("earlybirds")):
        "Number of penguins"

        @render.text
        def count():
            return filtered_df().shape[0]

    with ui.value_box(showcase=icon_svg("ruler-horizontal")):
        "Average bill length"

        @render.text
        def bill_length():
            return f"{filtered_df()['bill_length_mm'].mean():.1f} mm"

    with ui.value_box(showcase=icon_svg("ruler-vertical")):
        "Average bill depth"

        @render.text
        def bill_depth():
            return f"{filtered_df()['bill_depth_mm'].mean():.1f} mm"


with ui.layout_columns():
    with ui.card(full_screen=True):
        ui.card_header("Bill length and depth")

        @render.plot
        def length_depth():
            return sns.scatterplot(
                data=filtered_df(),
                x="bill_length_mm",
                y="bill_depth_mm",
                hue="species",
            )

    with ui.card(full_screen=True):
        ui.card_header("Penguin data")

        @render.data_frame
        def summary_statistics():
            cols = [
                "species",
                "island",
                "bill_length_mm",
                "bill_depth_mm",
                "body_mass_g",
            ]
            return render.DataGrid(filtered_df()[cols], filters=True)


ui.include_css(app_dir / "styles.css")


@reactive.calc
def filtered_df():
    filt_df = df[df["species"].isin(input.species())]
    filt_df = filt_df.loc[filt_df["body_mass_g"] < input.mass()]
    return filt_df



from shiny import App, Inputs, Outputs, Session, render, ui

app_ui = ui.page_fluid(
    ui.input_text_area(
        "caption_regular",
        "Caption:",
        "Data summary\nwith\nmultiple\nlines",
    ),
    ui.output_text_verbatim("value_regular", placeholder=True),
    ui.input_text_area(
        "caption_autoresize",
        ui.markdown("Caption (w/ `autoresize=True`):"),
        "Data summary\nwith\nmultiple\nlines",
        autoresize=True,
    ),
    ui.output_text_verbatim("value_autoresize", placeholder=True),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.text
    def value_regular():
        return input.caption_regular()

    @render.text
    def value_autoresize():
        return input.caption_autoresize()


app = App(app_ui, server)



# ------------------------------------------------------------------------------------
# A basic Shiny Chat example powered by OpenAI running on Azure.
# To run it, you'll need OpenAI API key.
# To get setup, follow the instructions at https://learn.microsoft.com/en-us/azure/ai-services/openai/quickstart?tabs=command-line%2Cpython-new&pivots=programming-language-python#create-a-new-python-application
# ------------------------------------------------------------------------------------
import os

from app_utils import load_dotenv
from openai import AzureOpenAI

from shiny.express import ui

# Either explicitly set the AZURE_OPENAI_API_KEY and AZURE_OPENAI_ENDPOINT environment
# variables before launching the app, or set them in a file named `.env`. The
# `python-dotenv` package will load `.env` as environment variables which can later be
# read by `os.getenv()`.
load_dotenv()

llm = AzureOpenAI(
    api_key=os.getenv("AZURE_OPENAI_API_KEY"),
    api_version="2024-02-01",
    azure_endpoint=os.getenv("AZURE_OPENAI_ENDPOINT"),  # type: ignore
)

deployment_name = "REPLACE_WITH_YOUR_DEPLOYMENT_NAME"

# Set some Shiny page options
ui.page_opts(
    title="Hello OpenAI Chat",
    fillable=True,
    fillable_mobile=True,
)

# Create a chat instance, with an initial message
chat = ui.Chat(
    id="chat",
    messages=[
        {"content": "Hello! How can I help you today?", "role": "assistant"},
    ],
)

# Display the chat
chat.ui()


# Define a callback to run when the user submits a message
@chat.on_user_submit
async def _():
    # Get messages currently in the chat
    messages = chat.messages(format="openai")
    # Create a response message stream
    response = llm.chat.completions.create(
        model=deployment_name,
        messages=messages,
        stream=True,
    )
    # Append the response stream into the chat
    await chat.append_message_stream(response)



import asyncio
import random
import sqlite3
from datetime import datetime
from typing import Any, Awaitable

import pandas as pd

from shiny import reactive
from shiny.express import input, render, ui

SYMBOLS = ["AAA", "BBB", "CCC", "DDD", "EEE", "FFF"]


def timestamp() -> str:
    return datetime.now().strftime("%x %X")


def rand_price() -> float:
    return round(random.random() * 250, 2)


# === Initialize the database =========================================


def init_db(con: sqlite3.Connection) -> None:
    cur = con.cursor()
    try:
        cur.executescript(
            """
            CREATE TABLE stock_quotes (timestamp text, symbol text, price real);
            CREATE INDEX idx_timestamp ON stock_quotes (timestamp);
            """
        )
        cur.executemany(
            "INSERT INTO stock_quotes (timestamp, symbol, price) VALUES (?, ?, ?)",
            [(timestamp(), symbol, rand_price()) for symbol in SYMBOLS],
        )
        con.commit()
    finally:
        cur.close()


conn = sqlite3.connect(":memory:")
init_db(conn)


# === Randomly update the database with an asyncio.task ==============


def update_db(con: sqlite3.Connection) -> None:
    """Update a single stock price entry at random"""

    cur = con.cursor()
    try:
        sym = SYMBOLS[random.randint(0, len(SYMBOLS) - 1)]
        print(f"Updating {sym}")
        cur.execute(
            "UPDATE stock_quotes SET timestamp = ?, price = ? WHERE symbol = ?",
            (timestamp(), rand_price(), sym),
        )
        con.commit()
    finally:
        cur.close()


async def update_db_task(con: sqlite3.Connection) -> Awaitable[None]:
    """Task that alternates between sleeping and updating prices"""
    while True:
        await asyncio.sleep(random.random() * 1.5)
        update_db(con)


_ = asyncio.create_task(update_db_task(conn))


# === Create the reactive.poll object ===============================


def tbl_last_modified() -> Any:
    df = pd.read_sql_query("SELECT MAX(timestamp) AS timestamp FROM stock_quotes", conn)
    return df["timestamp"].to_list()


@reactive.poll(tbl_last_modified, 0.5)
def stock_quotes() -> pd.DataFrame:
    return pd.read_sql_query("SELECT timestamp, symbol, price FROM stock_quotes", conn)


with ui.card():
    ui.markdown(
        """
        # `shiny.reactive.poll` demo

        This example app shows how to stream results from a database (in this
        case, an in-memory sqlite3) with the help of `shiny.reactive.poll`.
        """
    )
    ui.input_selectize("symbols", "Filter by symbol", [""] + SYMBOLS, multiple=True)

    @render.data_frame
    def table():
        df = stock_quotes()
        if input.symbols():
            df = df[df["symbol"].isin(input.symbols())]
        return df
